# coding: utf-8

import argparse
import os

from radicl import __version__
from radicl.interface import RADICL
from radicl.ui_tools import Messages

debug = False

# Get the screen outputter
out = Messages()


def main():
    logo = 'RADICL - RAD INTERFACE TO THE COMMANDLINE'
    hdr = '=' * len(logo)
    hdr = logo + '\n' + hdr
    parser = argparse.ArgumentParser(
        description="CLI script for live interactions with the Lyte probe.")
    parser.add_argument('--version', action='version',
                        version=('%(prog)s {version}'
                                 '').format(version=__version__))
    parser.add_argument('-d', '--debug', action='store_true', help='Log debug statements')
    args = parser.parse_args()

    # Clear the screen in console
    if 'nt' in os.name:
        os.system('cls')  # for Windows
    else:
        os.system('clear')  # for Linux/OS X

    out.msg("")
    out.msg(hdr)

    # Information
    out.headline("\nWELCOME TO THE LYTE PROBE CLI")
    out.warn("\n\tWarning: This CLI is not meant to run with the mobile app."
             "\n\tPlease make sure your RAD app is closed out.")
    out.msg("\nThings you can do with this tool:\n", 'header')
    out.msg("\t* Plot various data from the probe." +
            "\n\t* Write various data to a file. (In development)" +
            "\n\t* Modify probe settings.  (In development)" +
            "\n\t* Update the firmware (In development)\n")

    # try:
    cli = RADICL(debug=args.debug)
    cli.run()


if __name__ == '__main__':
    main()



from radicl.com import find_kw_port, get_serial_cnx, RAD_Serial
import pytest
from unittest.mock import patch
from types import SimpleNamespace


@pytest.mark.parametrize('kw, description_list, expected', [
    # Check plain usage with capital letters
    ('test', ['TEST device', 'unknown com1'], ['TEST device']),
    # check no match found
    ('test', ['com'], []),
    # Check multiple matches found
    ('test', ['TEST device', 'tester com1'], ['TEST device', 'tester com1']),

])
def test_find_kw_port(kw, description_list, expected):
    """
    Mock devices using a list of descriptions. Check the port finder
    grabs them
    """
    coms = [SimpleNamespace(description=c) for c in description_list]
    with patch('serial.tools.list_ports.comports', return_value=coms):
        result = find_kw_port(kw)
        assert [r.description for r in result] == expected


@pytest.mark.parametrize('kw, com_list, match_index, expected', [
    # One probe
    ('ST Micro', ['St Microelectronics', 'unknown'], 0, 'dev_fake_0'),
    # Two probes but pick the index
    ('ST Micro', ['St Microelectronics', 'St Microelectronics 1'], 1, 'dev_fake_1'),
    # No probe
    ('ST Micro', ['unknown 1', 'unknown 2'], 0, None),

])
def test_get_serial_cnx(kw, com_list, match_index, expected):
    devices = [SimpleNamespace(device=f'dev_fake_{i}', description=d) for i,d in enumerate(com_list)]
    with patch('serial.tools.list_ports.comports', return_value=devices):
        with patch('serial.Serial.open', return_value=None):
            cnx = get_serial_cnx(kw, match_index=match_index)
        if hasattr(cnx, 'port'):
            assert cnx.port == expected
        else:
            assert cnx == expected


class MockSerialPort:
    def __init__(self, **kwargs):
        self.port = kwargs['port']

    def setDTR(self, num):
        pass

    def close(self):
        pass

    def reset_output_buffer(self):
        pass

    def write(self, data):
        return len(data)

    def read(self):
        pass

    def inWaiting(self):
        pass


class TestRAD_Serial:

    @pytest.fixture(scope='function')
    def rs(self):
        with patch('serial.tools.list_ports.comports', return_value=[SimpleNamespace(device='dev_mock', description='STMicroelectronics')]):
            with patch('serial.Serial', return_value=MockSerialPort(port='mock')):
                rs = RAD_Serial()
                yield rs

    def test_rad_serial_open_port(self, rs):
        rs.openPort()
        assert rs.serial_port.port == 'mock'

    def test_closePort(self, rs):
        rs.closePort()
        assert rs.serial_port is None

    def test_flushPort(self, rs):
        rs.flushPort()

    def test_writePort(self, rs):
        n = rs.writePort(10)

    def test_writePortClean(self, rs):
        rs.writePortClean(10)

    def test_readPort(self, rs):
        rs.readPort()

    def test_numBytesInBuffer(self, rs):
        rs.numBytesInBuffer()



from enum import Enum


class CMDEnum(Enum):
    @property
    def cmd(self):
        return self.value


class AttributeCMD(CMDEnum):
    """Commands to retrieve non-measurement attributes """
    SERIAL = 0x04
    HW_ID = 0x01
    HW_REV = 0x02
    FW_REV = 0x03
    FULL_FW_REV = 0x09


class MeasCMD(CMDEnum):
    """Commands for measurements"""
    START = 0x42
    STOP = 0x43
    STATE = 0x40
    RESET = 0x41
    NUM_SEGMENTS = 0x44
    DATA_SEGMENT = 0x45
    TEMP = 0x4F


class SettingsCMD(CMDEnum):
    """Commands for probe settings"""

    SAMPLING_RATE = 0x46
    ZPFO = 0x47
    PPMM = 0x48
    ALG = 0x49
    APPP = 0x4A
    TCM = 0x4B
    USERTEMP = 0x4C
    IR = 0x4D
    CALIBDATA = 0x4E
    ACCTHRESH = 0x50
    ACCZPFO = 0x51
    ACCRANGE = 0x52


class SystemCMD(CMDEnum):
    """Commands for the system"""
    START_BOOTLOADER = 0x05
    STATUS = 0x06
    STATE = 0x07  # Run state


class FWUpdateCMD(CMDEnum):
    """Commands for the firmware"""
    ENTER = 0xF0
    STATE = 0xF1
    SIZE = 0xF2
    DOWNLOAD = 0xF3
    CRC = 0xF4
    CLOSE = 0xF5



# coding: utf-8

import logging
import sys
import textwrap
import coloredlogs
from colorama import init
from termcolor import colored
from threading import Timer


class Messages:
    init()

    def msg(self, str_msg, context_str=None):
        msg = colored(str_msg, 'white')
        print('\n' + msg)

    def dbg(self, str_msg):
        final_msg = colored('[DEBUG]: ', 'magenta', attrs=['bold'])
        final_msg += colored(str_msg, 'white')
        print('\n' + final_msg)

    def warn(self, str_msg):
        final_msg = colored('WARNING: ', 'yellow', attrs=['bold'])
        final_msg += colored(str_msg, 'white')
        print('\n' + final_msg)

    def error(self, str_msg):
        final_msg = colored('ERROR: ', 'red', attrs=['bold'])
        final_msg += colored(str_msg, 'red')
        print('\n' + final_msg)

    def critical(self, str_msg):
        final_msg = colored('ERROR: ', 'red', attrs=['bold'])
        final_msg += colored(str_msg + '\nExiting program...', 'red')
        print('\n' + final_msg)
        sys.exit()

    def respond(self, str_msg):
        final_msg = colored(str_msg, 'green')
        print('\t' + final_msg)

    def headline(self, str_msg):
        str_msg = str_msg.upper()
        final_msg = colored(str_msg, 'blue', attrs=['bold', 'underline'])
        print('\t' + final_msg)


def parse_func_list(func_list, identify_lst, ignore_keywords=[]):
    """
    Reads through a list of available methods and retrieves the methods that have
    matching words in the identify list, then it removes those keywords and Returns
    a dictionary of simplified method names with values as their corresponding
    method

    Args:
        func_list: A list of function names and functions
        identify_lst: A list of keywords to match with
        ignore_keywords: List contain keywords that will ignore functions with that keyword
    Returns:
        options: A dictionary that has keys that are simplified names of the methods, and values are the methods
    """
    options = {}
    ignore_keywords = [w.lower() for w in ignore_keywords]

    for f_name, fn in func_list:
        # Return a list of true when a word is found
        words_found = [True for w in identify_lst if w.lower()
                       in f_name.lower()]
        # If the number of the matches the number of keywords were
        # looking for
        if len(words_found) == len(identify_lst) and 'RAD' not in f_name:
            # Remove the keywords to form the simplified method name
            name = f_name
            for word in identify_lst:
                # print(f_name,word)
                name = name.replace(word, '')
            ignores = [True for ign in ignore_keywords if ign in name.lower()]
            if len(ignores) == 0:
                options[name.lower()] = fn

    return options


def parse_help(help_str):
    """
    Auto parses the help dialogs provided by developers in the doc strings.
    This is denoted by helpme - text and then a blank line.

    e.g.

    Function doc string
    helpme - more end user-friendly message about what this will do

    my arguments...

    """
    result = None

    if help_str is not None:
        if 'helpme' in help_str:
            z = help_str.split('helpme')
            result = z[-1].split('-')[-1].strip()

    return result


def print_helpme(help_str, help_dict):
    """
    Tries to help the user on request for help.

    Args:
        help_str -
        help_dict -
    """
    out_str = ""
    show_all = False
    no_help = 0
    no_doc = False

    # User provides specific help request
    if help_dict is not None:
        if "-" in help_str:
            h = help_str.split("-")
            k = ("".join(h[1:]).strip()).lower()
            if k in help_dict:
                formatted = textwrap.fill(long_str, width=width)

                for i, line in enumerate(formatted.split('\n')):
                    if i != 0:
                        k = ""
                    result += "\n{0:<20} {1:<20}\n".format(k, v)
            else:
                show_all = True
        else:
            show_all = True

        if show_all:
            for k, v in help_dict.items():
                if v is None:
                    no_help += 1
                    v = 'No help documentation.'

                out_str += '\n{0:<25} {1:<25}\n'.format(k, v)
        if no_help == len(help_dict.keys()):
            no_doc = True
    else:
        no_doc = True

    if no_doc:
        out_str = (
            '\nNo help documentation.\nPlease email info@adventuredata.com or file an issue at '
            'https://github.com/Adventuredata/radicl/issues\n')
        print(help_dict)

    # Doctor up the printout
    t = '\n{0:<20} {1:<20}\n'.format('OPTIONS', 'DESCRIPTION')
    print_able = '\n{0}'.format(colored(t, 'magenta', attrs=['bold']))
    print_able += out_str
    print(print_able)


def get_logger(name, debug=False, ext_logger=None):
    """
    Args:
        name: Name of the logger.
        debug: Bool whether to show debug statements
        ext_logger: Use ext logger to pass an already instantiated logger

    Returns:
        log: Instantiated logger
    """
    fmt = fmt = '%(name)s %(levelname)s %(message)s'
    if ext_logger is None:
        log = logging.getLogger(name)
    else:
        log = ext_logger
    if debug:
        level = 'DEBUG'
    else:
        level = 'INFO'

    coloredlogs.install(fmt=fmt, level=level, logger=log)
    return log


def exit_requested():
    ans = input('\nPress enter to begin listening for probe (type exit to quit): ')
    if ans.strip().lower() in ['exit', 'quit']:
        return True
    else:
        return False


def get_index_from_ratio(idx, ratio, n_samples):
    """
    Sometimes we have an index and we want something
    slightly near it. This manages grabbing those with
    consideration for the max value
    Args:
        idx: idx of interest
        ratio: scaling ratio (e.g. 1.1)
        n_samples: max allowable number of samples

    Returns:
        result: Scaled integer of the index
    """
    result = int(idx*ratio)
    if result > n_samples:
        result = n_samples-1
    return result



# coding: utf-8

import datetime
import inspect
import struct
import sys
import time

from . import __version__
from .com import RAD_Serial
from .api import RAD_API
from .ui_tools import get_logger, parse_func_list

error_codes = {2049: "The probe measurement/sensor is not running",
               2048: 'Generic Measurement error',
               2050: "Measurement FSM not ready to start measurement",
               2051: " Measurement FSM already stopped (i.e. in ready/idle state)",
               2052: "Setting a parameter failed",
               2053: "Invalid buffer was addressed",
               2054: "Reading data failed"}

# From Data sheet for accelerometer in REV C
acc_sensitivity = {2: 0.06,
                   4: 0.12,
                   6: 0.18,
                   8: 0.24,
                   16: 0.73}


class RAD_Probe:
    """
    Class for directly interacting with the probe.

    """
    __data_buffer_guide = ['Raw Sensor', 'Acceleration', ' Raw Pressure',
                           'Raw Depth', 'Filtered Depth',
                           'Acceleration Correlation',
                           'Pressure/Depth Correlation',
                           'Depth Corrected Sensor']

    def __init__(self, ext_api=None, debug=False):
        """
        Args:
            ext_api: rad_api.RAD_API object preinstantiated
        """

        self.log = get_logger(__name__, debug=debug)

        # Check if an external API object was passed in.
        if ext_api is not None:
            # An external API object was provided. Use it. Note: We assume here
            # that the appropriate initialization, identification, and API port
            # enable procedure has already taken place
            self.api = ext_api
        else:
            # No external API object was provided. Create new serial and API
            # objects for internal use
            port = RAD_Serial(debug=debug)
            port.openPort()

            if not port:
                self.log.info("No device present")
            else:
                port.flushPort()
                # Create the API and FMTR instances The API class is linked to
                # the port object
                api = RAD_API(port, debug=debug)

                # Switch the device over to API mode
                api.sendApiPortEnable()
                self.api = api

                # Delay a bit and then identify the attached device
                time.sleep(0.5)
                ret = api.Identify()

                if ret == 0:
                    self.log.error("Unable to connect to the probe. Unplug and"
                                   " power cycle it.")
                    sys.exit()
                time.sleep(0.1)

                # Manages the settings
                settings_funcs = inspect.getmembers(self.api,
                                                    predicate=inspect.ismethod)
                ignores = ['reset']
                self.settings = parse_func_list(settings_funcs,
                                                ['Meas', 'Set'],
                                                ignore_keywords=ignores)
                self.getters = parse_func_list(settings_funcs,
                                               ['Meas', 'Get'],
                                               ignore_keywords=ignores)

    def manage_error(self, ret_dict, stack_id=1):
        """
        Handles the common scenario of looking at the returned Dictionary
        from the probe where there may be an error or simply a com error.
        This function reports the name of the function and reports the error
        function.

        Args:
            ret_dict: Dictionary of keys ['Status','data','errorCode']
            stack_id: number of functions up the stack to use for reporting
                      function name when errors occur.
        """
        name = inspect.stack()[stack_id][3]

        if ret_dict['errorCode'] is not None:
            self.log.error("{} error:{}".format(name, ret_dict['errorCode']))

        else:
            self.log.error("{} error: COM".format(name))

    def manage_data_return(self, ret, num_values=1, dtype=int):
        """
        Manage a return from the probe when were expecting an integer

        Args:
            ret: Dictionary of the return with keys status, errorCode, and data
            num_values: Number of expect values to be returned
            dtype: Data type expected back

        Returns:
            result: None when no data was received. Integer when its not.
        """
        result = []

        if (ret['status'] == 1) and (ret['data'] is not None):

            data_size = len(ret['data'])
            increment = int(data_size / num_values)

            if data_size % num_values != 0:
                raise IOError("Data received (length = {}) is not evenly "
                              "divided by the expected number of values = "
                              "{}".format(data_size, num_values))

            for i in range(0, data_size, increment):
                data = ret['data'][i:i + increment]
                if dtype == int:
                    value = int.from_bytes(data, byteorder='little')

                elif dtype == str:
                    value = data.decode('utf-8')

                elif dtype == 'hex':
                    value = data.hex()

                else:
                    raise ValueError(
                        "No types other than str, hex, and int are implemented")

                result.append(value)

            if num_values == 1:
                result = result[0]

        else:
            self.manage_error(ret, stack_id=2)
            result = None

        return result

    def __readData(self, buffer_id, max_retry=10, init_delay=0.004):
        """
        Private function to retrieve data from the probe.
         Args:
            buffer_id: Integer specifying location in the probe buffer
            max_retry: Integer number of attempts before exiting with a fail
        """

        result = False
        num_segments = 0
        data = []

        buffer_name = self.__data_buffer_guide[buffer_id]
        self.log.info("Querying probe for {} data...".format(buffer_name.lower()))

        # Final data to return
        final = {'status': 0, 'SegmentsAvailable': 0, 'SegmentsRead': 0,
                 'BytesRead': 0,
                 'data': None}

        # Get the number of data segments available
        ret = self.api.MeasGetNumSegments(buffer_id)

        if ret['status'] == 1:
            num_segments = int.from_bytes(ret['data'], byteorder='little')
            self.log.debug('Retrieving {:,} segments of {} data...'.format(num_segments, buffer_name.lower()))

        # No data returned
        else:

            self.log.debug("getNumSegments error: %d (buffer_id = %d)" %
                           (ret['errorCode'], buffer_id))

            if ret['errorCode'] is not None:
                self.log.debug("Error {:d} occurred.".format(ret['errorCode']))

                if ret['errorCode'] in error_codes.keys():
                    self.log.error(error_codes[ret['errorCode']])
                else:
                    self.log.warning("Unknown error code!")

        # If we do have number of segments
        if num_segments != 0 and num_segments is not None:
            self.log.debug("Reading %d segments" % num_segments)
            byte_counter = 0

            # Data Segments to collect
            for ii in range(0, num_segments):
                result = False

                # initial delay time
                wait_time = init_delay

                # Delays and retry
                for jj in range(0, max_retry):

                    time.sleep(wait_time)

                    # Request the data
                    ret = self.api.MeasReadDataSegment(buffer_id, ii)

                    if ret['status'] == 1 and ret['data'] is not None:
                        byte_counter = byte_counter + len(ret['data'])
                        data_chunk = ret['data']
                        data.extend(data_chunk)
                        result = True
                        # Break the retry loop
                        break

                    else:
                        if jj >= max_retry:
                            self.log.warning('Missed data segment, after {0:d} attempts.'.format(jj + 1))

                        # Developer friendly response in event of read error
                        msg = ("{0} Data Error: Buffer ID = {1:d}, "
                               "Segment ID={2:d}/{3:d}, Retry #{4:d}, "
                               " COM Delay = {5}s "
                               "").format(buffer_name, buffer_id, ii,
                                          num_segments, jj, wait_time)

                        self.log.debug(msg)

                        # Increase the wait time every failed request
                        wait_time += wait_time

                        if ret['errorCode'] is not None:
                            if ret['errorCode'] in error_codes.keys():
                                self.log.warning("Received error code:{}".format(error_codes[ret['errorCode']]))
                            else:
                                self.log.warning("Received unknown error code:{}".format(ret['errorCode']))

            # Was the data read successful?
            final['status'] = int(result)
            final['SegmentsAvailable'] = num_segments
            final['SegmentsRead'] = ii + 1  # Report this not zero based
            final['BytesRead'] = byte_counter

            if final['SegmentsRead'] > 0:
                final['data'] = data

        return final

    # ********************
    # * PUBLIC FUNCTIONS *
    # ********************
    def getProbeSerial(self):
        """
        Returns the probe's serial number. The return value is a string. If the
        request fails it will return None
        """

        ret = self.api.getSerialNumber()
        if ret['data'] is not None:
            # Flip the byte array since it comes in backwards
            ret['data'] = ret['data'][::-1]
            # upper case to match the store serial number lists
            return self.manage_data_return(ret, dtype='hex').upper()
        else:
            return self.manage_data_return(ret, dtype='hex')

    def getProbeSystemStatus(self):
        """
        Returns the probe's system status. The return value is an integer.
        If the request fails it will return None
        """

        ret = self.api.getSystemStatus()

        return self.manage_data_return(ret, dtype=int)

    def getProbeRunState(self):
        """
        Returns the probe's run state. The return value is an integer.
        If the request fails it will return None
        """

        ret = self.api.getRunState()
        return self.manage_data_return(ret, dtype=int)

    def getProbeMeasState(self):
        """
        Retrieves the probe measurement state and converts it to an integer

        Returns:
            integer-measurement state of the probe, or none if error arises.
        """
        data = None
        attempts = 0

        while data is None and attempts < 10:
            ret = self.api.getMeasState()
            data = self.manage_data_return(ret, dtype=int)
            attempts += 1

        return data

    def startMeasurement(self):
        """
        Starts a new measurement. Returns 1 if successful, 0 otherwise
        """

        ret = self.api.MeasStart()
        self.log.debug("Start measurement requested.")

        if ret['status'] == 1:
            self.wait_for_state(1)
            self.log.info("Measurement started...")

            return 1

        else:
            self.manage_error(ret)

            return 0

    def stopMeasurement(self):
        """
        Stops an ongoing measurement. Returns 1 if successful, 0 otherwise
        """

        ret = self.api.MeasStop()
        self.log.debug("Stop measurement requested.")

        if ret['status'] == 1:
            self.wait_for_state(3)
            self.log.info("Measurement stopped...")

            return 1
        else:
            self.manage_error(ret)

            return 0

    def resetMeasurement(self):
        """
        Resets the measurement FSM to prepare for a new measurement.
        Returns 1 if successful, 0 otherwise
        """

        ret = self.api.MeasReset()
        self.log.debug("Measurement reset requested...")

        if ret['status'] == 1:
            self.wait_for_state(0, delay=0.1)
            self.log.info("Probe measurement reset...")

            return 1

        else:
            self.manage_error(ret)

            return 0

    def wait_for_state(self, state, retry=500, delay=0.2):
        """
        Waits for the specified state to occur. This is particularly useful when
        a command is requested.

        Args:
            state: single integer
            retry: Number of attempts to try while Waiting for the states
            delay: time in seconds to wait between each attempt

        """

        attempts = 0
        pstate = self.getProbeMeasState()
        result = False
        self.log.info(
            "Waiting for state {0}, current state = {1}".format(
                state, pstate))

        while not result:
            result = pstate == state

            # Check for a probe advanced past the state
            if isinstance(pstate, int):
                if state != 0:
                    if pstate >= state:
                        result = True
                if state == 5:
                    if pstate == 0:
                        result = True

            if attempts > retry:
                self.log.error(
                    "Retry Exceeded waiting for state(s) {0}".format(state))
                result = False
                break
            else:
                attempts += 1

            time.sleep(delay)
            pstate = self.getProbeMeasState()

        if result:
            self.log.debug(
                "{} queries while waiting for state {}".format(
                    attempts, state))

        return result

    def read_check_data_integrity(self, buffer_id, nbytes_per_value=None,
                                  nvalues=None, from_spi=False):
        """
        Receives a data function and  performs the data integrity check
        If the data is from _spi then we know how long the segments are.
        If there are not, then it is possible we receive an incomplete segment
        so we check for integer_multiples of that data

        Args:
            buffer_id: Index the data is in the probe buffer
            nbytes_per_value: how many bytes expected per value per sample
            nvalues: Number of values expecter per sample
            from_spi: Is the data coming from SPI flash which guarantees a
                      segment size
        Returns:
            final: dict containing data in bytes, number of samples and
                   segements and return status.
        """

        buffer_name = self.__data_buffer_guide[buffer_id]

        ret_dict = self.__readData(buffer_id)
        final = None

        # successfully read data
        if ret_dict['status'] != 1:
            self.log.error('Read {} error: No data available!'
                           ''.format(buffer_name))
            self.log.debug('Data received from probe:\n{}'.format(ret_dict))

        else:

            # ***** DATA INTEGRITY CHECK *****
            # Check for all segments read in
            all_segments = (ret_dict['SegmentsAvailable'] == \
                            ret_dict['SegmentsRead'])
            int_multiple = nbytes_per_value * nvalues

            # Grab the number of samples, NOTE this is only valid if the
            # ...conditions are true below
            samples = ret_dict['BytesRead'] // int_multiple

            # Data from SPI Flash
            if from_spi:
                expected_bytes = ret_dict['SegmentsRead'] * 256

                # Check we read all bytes:
                complete_bytes = expected_bytes == ret_dict['BytesRead']
                self.log.debug('Downloaded {:0,.2f}/{:0,.2f} Kb.'.format(
                    ret_dict['BytesRead'] / 1000,
                    expected_bytes / 1000))
            # From chip memory
            else:
                # We can have incomplete segments, so check for even numbers
                complete_bytes = ret_dict['BytesRead'] % int_multiple == 0
                self.log.debug('Byte Multiples: {:0,.2f}'.format(ret_dict['BytesRead'] / int_multiple))

            # Check the data integrity
            if not all_segments or not complete_bytes:
                self.log.error("Data Integrity Error: Unable to retrieve all "
                               "data for {}.".format(buffer_name))
                self.log.debug('All Segments not downloaded: {}, All bytes '
                               'downloaded: {}'.format(all_segments, complete_bytes))
            else:
                final = ret_dict
                final['samples'] = samples

                # Final reporting
                self.log.info('Retrieving {:,} samples of {} data...'
                              ''.format(ret_dict['samples'], buffer_name))

            self.log.debug("Segment Retrieved {:,d}/{:,d}."
                           " Bytes Retrieved {:0,.2f} Kb"
                           "".format(ret_dict['SegmentsRead'],
                                     ret_dict['SegmentsAvailable'],
                                     ret_dict['BytesRead'] / 1000))

            return final

    def readRawSensorData(self):
        """
        Reads the RAW sensor data.
        helpme - Raw NIR data for 4 sensors

        Returns:
            dict - containing data or None if read failed
        """

        # Index in the probe buffer
        buffer_id = 0

        # Expected bit size of the data
        sgbytes = 256
        nbytes_per_value = 2
        nvalues = 4

        ret = self.read_check_data_integrity(buffer_id,
                                             nbytes_per_value=nbytes_per_value,
                                             nvalues=nvalues, from_spi=True)

        # ***** DATA PARSING *****
        if ret is not None:
            data = ret['data']
            samples = ret['samples']
            offset = 0
            final = {'Sensor1': [], 'Sensor2': [], 'Sensor3': [], 'Sensor4': []}

            for ii in range(0, samples):

                # Loop over each sensor and add it to the final dict
                for idx in range(nvalues):
                    # Create a name for the dictionary
                    name = 'Sensor{}'.format(idx + 1)

                    # Form the index for the bytes for the start of each value.
                    byte_idx = idx * nbytes_per_value + offset

                    # Grab each sensor value in the byte array
                    final[name].append(data[byte_idx] + \
                                       data[byte_idx + 1] * sgbytes)

                # Move farther down the bytes to read the next segment.
                offset += nbytes_per_value * nvalues

            data = final
        return data

    def readCalibratedSensorData(self):
        """
        Reads the Read Calibrated sensor data.
        helpme - Calibrated data for 4 sensors

        Returns:
            dict - containing data or None if read failed
        """
        calib_data = {}
        raw = self.readRawSensorData()

        for id in range(1, 5):
            sensor = "Sensor{}".format(id)
            d = self.getSetting(setting_name='calibdata', sensor=id)

            # For the calibration for sensor 1 we invert
            if id == 1:
                # Set the slope to the negative difference
                m = 4095 / (d[0] - d[1])
                # Set the intercept to the LOW value
                b = d[1]

            else:
                # Set the slope to the positive difference
                m = 4095 / (d[1] - d[0])
                # Set the intercept to the LOW value
                b = d[0]

            calib_data[sensor] = [m * (x - b) for x in raw[sensor]]

        return calib_data

    def readRawAccelerationData(self):
        """
        Reads the raw 3 axis  acceleration data
        helpme - Raw 3 axis accelerometer data from the probe

        Returns:
            dict: containing accel data (x,y,z) or None if read failed

        """
        acc_axes = ['X', 'Y', 'Z']
        name_str = '{}-Axis'
        # Index in the probe buffer
        buffer_id = 1

        # Expected bit size of the data
        nbytes_per_value = 2
        nvalues = len(acc_axes)

        ret = self.read_check_data_integrity(buffer_id,
                                             nbytes_per_value=nbytes_per_value,
                                             nvalues=nvalues)
        data = None

        if ret is not None:
            data = ret['data']

            samples = ret['samples']

            # Initialize the data
            final = {name_str.format(a): [] for a in acc_axes}
            offset = 0

            # Grab the range to scale the incoming data
            a_r = self.getSetting(setting_name='accrange')
            sensitivity = acc_sensitivity[a_r]

            # Loop over all the samples
            for ii in range(0, samples):
                # Loop over the sample and extract each axis
                for idx, axis in enumerate(acc_axes):
                    # Form the dictionary name
                    name = name_str.format(axis)

                    # Form the index of where each axis is in the sample
                    byte_idx = idx * nbytes_per_value + offset

                    # Grab the bytes from the list for a single axis
                    byte_list = data[byte_idx: (byte_idx + nbytes_per_value)]

                    # Form a byte object
                    byte_object = bytes(byte_list)

                    # Unpack bytes originally a 32 bit long and save
                    value = struct.unpack('<h', byte_object)[0]

                    # Convert to milli-gs while accounting for acc. range
                    value *= sensitivity
                    # Convert to Gs.
                    value /= 1000

                    final[name].append(value)

                # Advance forward in the data from probe for parsing
                offset += nbytes_per_value * nvalues

            # Assign final to the original data var for return
            data = final

        return data

    def readAccelerationCorrelationData(self):
        """
        Reads the acceleration correlation data
        """

        ret = self.__readData(5)
        if ret['status'] == 1:
            # successfully read data
            # ***** DATA INTEGRITY CHECK *****
            # Data integrity error (not all segments read)
            if ret['SegmentsAvailable'] != ret['SegmentsRead']:
                self.log.error("readAccelerationCorrelationData error: Data integrity "
                               "error (not all segments read)")
                return None

            total_bytes = ret['BytesRead']

            if (total_bytes % 4) != 0:
                # Data integrity error (not all bytes read - incomplete data segment)
                # The data set is not an integer multiple of 4 (Correlation
                # number is a 32-bit int => 4 bytes)
                self.log.error("readAccelerationCorrelationData error: Data integrity "
                               "error (incomplete data set)")
                return None

            # ***** DATA PARSING *****
            data = ret['data']
            correll_data = []
            samples = total_bytes // 4
            offset = 0

            for ii in range(0, samples):
                correll_data.append(data[(offset + 0)] + (data[(offset + 1)] * 256) + (
                    data[(offset + 2)] * 65536) + (data[(offset + 3)] * 16777216))
                offset = offset + 4
            return correll_data
        else:
            # Read failed!
            return None

    def readRawPressureData(self):
        """
        Reads the RAW pressure data, including the correlation index
        """

        ret = self.__readData(2)

        # successfully read data
        if ret['status'] == 1:
            # ***** DATA INTEGRITY CHECK *****
            if ret['SegmentsAvailable'] != ret['SegmentsRead']:
                # Data integrity error (not all segments read)
                self.log.error("readRawPressureData error: Data integrity error (not "
                               "all segments read)")
                return None

            # Data integrity error (not all bytes read - incomplete data
            # segment)
            total_bytes = ret['BytesRead']
            if (total_bytes % 3) != 0:
                # The data set is not an integer multiple of 3 (Raw pressure is
                # in 24-bit format => 3 bytes)
                self.log.error(
                    "readRawPressureData error: Data integrity error (incomplete data set)")
                return None

            # ***** DATA PARSING *****
            data = ret['data']
            pressure_data = []
            samples = total_bytes // 3
            offset = 0

            for ii in range(0, samples):
                this_value = data[(offset + 0)] + (data[(offset + 1)] * 256) + \
                             (data[(offset + 2)] * 65536)

                pressure_data.append(this_value / 4096)
                offset = offset + 3
            return pressure_data

        # Read failed!
        else:
            return None

    def readRawDepthData(self):
        """
        Reads the converted depth data, including the correlation index
        Return type is a dict containing data or None if read failed
        helpme - Unfiltered elevation data from the depth sensor
        """
        ret = self.__readData(3)
        if ret['status'] == 1:
            # successfully read data
            # ***** DATA INTEGRITY CHECK *****
            if ret['SegmentsAvailable'] != ret['SegmentsRead']:
                # Data integrity error (not all segments read)
                self.log.error("readDepthData error: Data integrity error (not all "
                               "segments read)")
                return None

            total_bytes = ret['BytesRead']
            if (total_bytes % 4) != 0:
                # Data integrity error (not all bytes read - incomplete data segment)
                # The data set is not an integer multiple of 4 (Floating point)
                # values are 32-bit long => 4 bytes)
                self.log.error("readDepthData error: Data integrity error "
                               "(incomplete data set)")
                return None

            # ***** DATA PARSING *****
            data = ret['data']
            depth_data = []
            samples = total_bytes // 4
            offset = 0
            for ii in range(0, samples):
                this_byte_list = data[(offset + 0):(offset + 4)]
                this_byte_object = bytes(this_byte_list)
                this_value = struct.unpack('f', this_byte_object)
                depth_data.append(this_value)
                offset = offset + 4
            return depth_data
        else:
            # Read failed!
            return None

    def readFilteredDepthData(self):
        """
        Retrieves the filtered depth data according to the zero-phase low-pass
        filter (equivalent to Matlab's 'filtfilt'). The amount of filtering
        applied is set by :func:`~radicl.RAD_API.MeasSetZPFO`

        helpme - The probe's filtered depth data is the raw depth data made
                 relative to the change and the amount filtered set by the ZPFO
                 option.

        """

        # Index in the probe buffer
        buffer_id = 4

        # Expected bit size of the data
        nbytes_per_value = 4
        nvalues = 1

        ret = self.read_check_data_integrity(buffer_id,
                                             nbytes_per_value=nbytes_per_value,
                                             nvalues=nvalues)
        data = None

        if ret is not None:
            data = ret['data']
            samples = ret['samples']
            depth_data = []
            offset = 0

            for ii in range(0, samples):
                this_byte_list = data[offset:(offset + nbytes_per_value)]
                this_byte_object = bytes(this_byte_list)
                this_value = struct.unpack('f', this_byte_object)
                depth_data.append(this_value)
                offset += nbytes_per_value

            # Convert to cm
            depth_data = [(c[0] / 100.0) for c in depth_data]
            data = depth_data

        return data

    def readPressureDepthCorrelationData(self):
        """
        Reads the pressure/depth correlation data
        """

        ret = self.__readData(6)
        if ret['status'] == 1:
            # successfully read data
            # ***** DATA INTEGRITY CHECK *****
            if ret['SegmentsAvailable'] != ret['SegmentsRead']:
                # Data integrity error (not all segments read)
                self.log.error("readPressureDepthCorrelationData error: Data integrity "
                               "error (not all segments read)")
                return None

            total_bytes = ret['BytesRead']
            if (total_bytes % 4) != 0:
                # Data integrity error (not all bytes read - incomplete data segment)
                # The data set is not an integer multiple of 4 (Correlation
                # number is a 32-bit int => 4 bytes)
                self.log.error("readPressureDepthCorrelationData error: Data integrity "
                               "error (incomplete data set)")
                return None

            # ***** DATA PARSING *****
            data = ret['data']
            correll_data = []
            samples = total_bytes // 4
            offset = 0

            for ii in range(0, samples):
                correll_data.append(data[(offset + 0)] + (data[(offset + 1)] *
                                                          256) + (data[(offset + 2)] * 65536) +
                                    (data[(offset + 3)] * 16777216))
                offset = offset + 4
            return correll_data

        else:
            # Read failed!
            return None

    def readDepthCorrectedSensorData(self):
        """
        Reads the processed data as a sensor-depth combo
        """
        # Data columns
        nsensors = 4
        name_str = 'Sensor{}'
        sensor_names = [name_str.format(i) for i in range(1, nsensors + 1)]

        # Index in the probe buffer
        buffer_id = 7

        # Expected byte size of the data
        nbytes_per_value = 4
        nvalues = 4

        ret = self.read_check_data_integrity(buffer_id,
                                             nbytes_per_value=nbytes_per_value,
                                             nvalues=nvalues, from_spi=False)
        data = None

        if ret is not None:
            # initialize
            data = ret['data']
            samples = ret['samples']
            final = {sensor: [] for sensor in sensor_names}
            final['depth'] = []
            nbytes = nbytes_per_value * nvalues
            sgbytes = 256
            offset = 0

            # Loop over all the samples
            for ii in range(0, samples):
                data_set = data[offset: offset + nbytes]

                # Grab the depth
                final['depth'].append(data_set[4] + (data_set[5] * sgbytes) +
                                      (data_set[6] * 65536) + (data_set[7] *
                                                               16777216))

                for idx, name in enumerate(sensor_names):
                    # Starts at idx 8 in the buffer
                    byte_idx = idx * 2 + 8
                    final[name].append(data_set[byte_idx] + \
                                       data_set[byte_idx + 1] * sgbytes)
                offset += nbytes

            data = final

        return data

    def readMeasurementTemperature(self):
        """
        Reads the temperature from the last measurement
        """

        ret = self.api.MeasGetMeasTemp()
        return self.manage_error(ret)

    def getProbeHeader(self):
        """
        Returns the lines that should be in every data file produced.
        """
        t = datetime.datetime.now()
        fstr = "{0}-{1:02d}-{2:02d}--{3:02d}:{4:02d}:{5:02d}"
        time_stamp = fstr.format(t.year, t.month, t.day,
                                 t.hour, t.minute, t.second)
        header = {"RECORDED": time_stamp,
                  "radicl VERSION": __version__,
                  "FIRMWARE REVISION": self.api.fw_rev,
                  "HARDWARE REVISION": self.api.hw_rev,
                  "MODEL NUMBER": self.api.hw_id}
        return header

    def getSetting(self, setting_name=None, sensor=None):
        """
        Reads the probes setting from the dictionary of functions. Calls the
        function and manages the data.

        Args:
            setting_name: name of the function minus Meas and get
            sensor: Sensor ID for calibration data

        Returns:
            int: from the function getting the probe setting, or list of 2 for calibration data
        """
        if setting_name == 'calibdata':
            ret = self.getters[setting_name](sensor)
            num_values = 2
        else:
            ret = self.getters[setting_name]()
            num_values = 1

        return self.manage_data_return(ret, num_values=num_values, dtype=int)

    def setSetting(self, setting_name=None, sensor=None, value=None, low_value=None,
                   hi_value=None):
        """
        sets the probe's setting
        """

        if setting_name == 'calibdata':
            ret = self.settings[setting_name](sensor, low_value, hi_value)

        else:
            ret = self.settings[setting_name](value)

        if ret['status'] == 1:
            return True

        else:
            self.manage_error(ret)

        return None



#!/usr/bin/env python

import argparse
import os
import sys
import platform
import time
import traceback
from matplotlib import pyplot as plt
import matplotlib
from radicl.ui_tools import get_logger, get_index_from_ratio

from study_lyte.detect import get_acceleration_start, get_acceleration_stop, get_nir_surface
from study_lyte.depth import get_depth_from_acceleration, get_constrained_baro_depth
from study_lyte.io import read_csv
from study_lyte.adjustments import remove_ambient


matplotlib.rcParams['agg.path.chunksize'] = 100000


def plot_events(ax, start=None, surface=None, stop=None, nir_stop=None, plot_type='normal'):
    """
    Plots the hline or vline for each event on a plot
    Args:
        ax: matplotlib.Axes to add horizontal or vertical lines to
        start: Array index to represent start of motion
        surface: Array index to represent snow surface
        stop: Array index to represent stop of motion
        nir_stop: Array index to represent stop estimated by nir
        plot_type: string indicating whether the index is on the y (vertical) or the x (normal)
    """
    event_alpha = 0.6
    if plot_type == 'vertical':
        line_fn = ax.axhline
    elif plot_type == 'normal':
        line_fn = ax.axvline
    else:
        raise ValueError(f'Unrecognized plot type {plot_type}, options are vertical or normal!')
    if start is not None:
        line_fn(start, linestyle='dashed', color='lime', label='Start', alpha=event_alpha)
    if surface is not None:
        line_fn(surface, linestyle='dashdot', color='cornflowerblue', label='Surface', alpha=event_alpha)
    if stop is not None:
        line_fn(stop, linestyle='dashed', color='red', label='Stop', alpha=event_alpha)
    if nir_stop is not None:
        line_fn(nir_stop, linestyle='dashed', color='magenta', label='NIR Stop', alpha=event_alpha)


def plot_hi_res(fname=None, df=None, timed_plot=None, calibration_dict={}):
    """
    Plots the timeseries, the depth corrected, accelerometer and depth data.
    Plot from a dataframe or from a file. Use auto close to auto close the figure
    after a some amount of time.

    Args:
        fname: Path to csv containing hi resolution data
        df: Optional pandas dataframe instead of a file
        timed_plot: Amount of time to show the plot, if none user has to close it
        calibration_dict: Dictionary to offer calibration coefficients for any of the sensors

    """
    log = get_logger('Hi Res Plot')
    if 'Linux' in platform.platform():
        matplotlib.use('TkAgg')

    print('')
    if fname is not None:
        df, meta = read_csv(fname)
        log.info(f"Filename: {fname}")

    # Setup a panel of plots
    fig = plt.figure(figsize=(10, 6), constrained_layout=True)
    gs = fig.add_gridspec(2, 5)

    # # Use time when possible
    if 'time' in df.columns:
        time_series = df['time']
    else:
        time_series = df.index

    # Grab all the estimates on the typical events of interest
    if 'Y-Axis' in df.columns:
        detect_col = 'Y-Axis'
        acc_cols = [c for c in df.columns if 'Axis' in c]

    elif 'acceleration' in df.columns:
        detect_col = 'acceleration'
        acc_cols = ['acceleration']
    else:
        raise ValueError('No acceleration was found in the file!')

    # Estimate events
    start = get_acceleration_start(df[detect_col])
    stop = get_acceleration_stop(df[detect_col])

    # Calculate depth from acceleration
    acc_depth = get_depth_from_acceleration(df[acc_cols + ['time']]).mul(100)
    acc_depth = acc_depth.reset_index()
    df['acc_depth'] = acc_depth[detect_col].copy()
    df['avg_depth'] = df[['acc_depth', 'depth']].mean(axis=1)

    # Crop data to motion
    cropped = df.iloc[start:stop].copy()
    cropped['clean'] = remove_ambient(cropped['Sensor3'], cropped['Sensor2'])

    surface = get_nir_surface(cropped['clean'])
    full_surface = surface + start

    # Re-zero the depth
    cropped['depth'] = cropped['depth'] - cropped['depth'].iloc[surface]
    cropped['acc_depth'] = cropped['acc_depth'] - cropped['acc_depth'].iloc[surface]
    cropped['avg_depth'] = cropped[['acc_depth', 'depth']].mean(axis=1)

    # Calculate some travel distances
    depth_cols = ['depth', 'acc_depth', 'avg_depth']
    max_distance = df[depth_cols].max() - df[depth_cols].min()
    mv_distance = df[depth_cols].iloc[start] - df[depth_cols].iloc[stop]
    snow_distance = df[depth_cols].iloc[full_surface] - df[depth_cols].iloc[stop]

    # print out some handy numbers
    log.info(f"* Number of samples: {len(df.index):,}\n")
    msg = '{:<25}{:<10}{:<10}{:<10}'
    header = msg.format('Distance', 'Baro.', 'Accel.', 'Avg')
    log.info(header)
    log.info('-' * len(header))
    msg = '{:<25}{:<10.1f}{:<10.1f}{:<10.1f}'
    distances = {"Maximum Measured": max_distance,
                 'During Motion': mv_distance,
                 'Snow Only': snow_distance}
    for desc, distance in distances.items():
        log.info(msg.format(desc, distance['depth'], distance['acc_depth'],
                            distance['avg_depth']))

    # Provide depth shifts
    ambient_shift = 6  # cm
    active_shift = 4.5  #
    mean_shift = (ambient_shift + active_shift) / 2
    time_series_events = dict(start=time_series[start],
                              surface=time_series[full_surface],
                              stop=time_series[stop],
                              nir_stop=None,
                              plot_type='vertical')

    # Plot time series data force data
    ax = fig.add_subplot(gs[:, 0])
    plot_events(ax, **time_series_events)
    ax.plot(df['Sensor1'], time_series, color='k')
    ax.set_title("Raw Force Timeseries")
    ax.legend(loc='lower left')
    ax.set_ylabel('Time [s]')
    ax.set_xlim(0, 4096)
    ax.invert_yaxis()

    # plot time series NIR data
    ax = fig.add_subplot(gs[:, 1])
    plot_events(ax, **time_series_events)
    ax.plot(df['Sensor2'], time_series, color='darkorange', label='Ambient')
    ax.plot(df['Sensor3'], time_series, color='crimson', label='Active')
    ax.set_title("NIR Timeseries")
    ax.legend(loc='lower left')
    ax.invert_yaxis()

    # plot the depth corrected Force
    ax = fig.add_subplot(gs[:, 2])
    ax.grid(True, axis='y', which='both', alpha=0.5)
    plot_events(ax, surface=cropped['acc_depth'].iloc[surface],
                plot_type='vertical')
    ax.plot(cropped['Sensor1'], cropped['acc_depth'], color='k', label='Raw Force')
    ax.set_title("Force Depth Corrected")
    ax.legend(loc='lower left', fontsize='small')
    ax.set_xlim((0, 4096))
    ax.set_ylabel('Depth [cm]')

    # plot depth corrected NIR data
    ax = fig.add_subplot(gs[:, 3])
    ax.grid(True, axis='y', alpha=0.5)
    plot_events(ax, surface=cropped['depth'].iloc[surface], plot_type='vertical')
    ax.plot(cropped['Sensor2'], cropped['acc_depth'] + 2.0, alpha=0.3, color='darkorange', label='Ambient')
    ax.plot(cropped['Sensor3'], cropped['acc_depth'], alpha=0.3, color='crimson', label='Active')
    ax.plot(cropped['clean'], cropped['acc_depth'], color='crimson', label='Clean Active')
    ax.set_title("NIR Depth Corrected")
    ax.legend(loc='lower left', fontsize='small')

    # plot the acceleration as a sub-panel with events, handle acceleration or all 3 axis
    ax = fig.add_subplot(gs[0, 4])
    # Switch event direction plotting for horiz. time series
    time_series_events["plot_type"] = 'normal'
    plot_events(ax, **time_series_events)
    acc_colors = ['darkslategrey', 'darkgreen', 'darkorange']
    for aidx, c in enumerate(acc_cols):
        ax.plot(time_series, df[c], color=acc_colors[aidx], label=c)
    ax.set_ylabel("Acceleration [g's]")
    ax.set_xlabel('Time [s]')
    ax.set_title('Accelerometer')
    ax.legend(fontsize='xx-small')

    # plot the depth as a sub-panel with events
    ax = fig.add_subplot(gs[1, 4])
    ax.set_title('Depth')
    plot_events(ax, **time_series_events)
    ax.plot(time_series, df['depth'], color='navy', label='Baro.')
    ax.plot(time_series, df['acc_depth'], color='mediumseagreen', label='Acc.')
    ax.plot(time_series, df['avg_depth'], color='tomato', label='Avg.')
    extra = get_constrained_baro_depth(df)
    ax.plot(extra.index, extra['depth'], label='constr.')
    ax.legend(loc='upper right', fontsize='xx-small')
    ax.set_ylabel('Depth from Max Height [cm]')
    # limits for depth
    buffer = 0.3
    n_samples = len(df.index)
    lim_idx1 = get_index_from_ratio(start, 1-buffer, n_samples)
    lim_idx2 = get_index_from_ratio(stop, 1 + buffer, n_samples)
    ts1, ts2 = time_series[lim_idx1], time_series[lim_idx2]
    depth1 = df[['depth', 'acc_depth']].iloc[lim_idx1].max() * (1-buffer)
    depth2 = df[['depth', 'acc_depth']].iloc[lim_idx2].min() * (1+buffer)
    if abs(depth1) < 5:
        depth1 = 5

    ax.set_xlim(ts1, ts2)
    ax.set_ylim(*sorted([depth1, depth2]))
    ax.grid(True, axis='y', which='both', alpha=0.5)

    # Make the figure full screen
    manager = plt.get_current_fig_manager()
    if 'Linux' in platform.platform():
        manager.full_screen_toggle()

    # Manage plot time
    if timed_plot == 0:
        pass
    elif timed_plot is not None:
        timer = fig.canvas.new_timer(interval=timed_plot*1000)
        timer.add_callback(plt.close)
        timer.start()
        plt.show(block=True)

    else:
        plt.show()


def plot_hi_res_cli():
    files = sys.argv[1:]

    for f in files:
        try:
            plot_hi_res(fname=f)
        except Exception as e:
            print(f'Encountered error while processing {f}')
            traceback.print_exc(file=sys.stdout)
            time.sleep(1)
            plt.close()


def main():
    parser = argparse.ArgumentParser(
        description='Plot various versions of probe data.')
    parser.add_argument('file', help='path to measurement', nargs='+')

    parser.add_argument('--smooth''-sm', dest='smooth',
                        help='Provide a integer to describing number of windows to smooth over')

    parser.add_argument('--avg', '-a', dest='average', action='store_true',
                        help='Average all three sensors together')
    parser.add_argument('--compare', '-c', dest='compare', action='store_true',
                        help='Plots before and after sensors')
    parser.add_argument('--sensor', '-sn', dest='sensor', type=int,
                        help='plots only a specific sensor, must be between 1-4')

    args = parser.parse_args()

    # Provide a opportunity to look at lots
    filenames = []
    if args.file is not None and not isinstance(args.file, list):
        if os.path.isdir(args.file):
            filenames = os.listdir(args.file)
        elif os.path.isfile(args.file):
            filenames = [args.file]

    elif isinstance(args.file, list):
        filenames = args.file
    else:
        print("Please provide a directory or filename")
        sys.exit()

    for f in filenames:
        try:
            post_processed = False
            print('\n' + os.path.basename(f))
            print('=' * len(f))
            # Open the file and set the index to depth
            df_o, meta = read_csv(f)
            df = df_o.copy()

            if args.smooth is not None:
                df = df.rolling(window=int(args.smooth)).mean()
                post_processed = True

            print("Pre-processed profile:")
            print("\tNum of samples: {0:,}".format(len(df_o.index)))
            print("\tDepth achieved: {0:.1f}".format(min(df_o.index)[-1]))
            print("\tResolution: {0:.1f} pts/cm".format(
                abs(len(df_o.index) / min(df_o.index)[1] - max(df_o.index)[1])))

            if post_processed:
                print("\nPost-processed profile:")
                print("\tNum of samples: {0:,}".format(len(df.index)))
                print("\tDepth achieved: {0:.1f}".format(min(df.index)))
                print(
                    "\tResolution: {0:.1f} pts/cm".format(abs(len(df.index) / min(df.index) - max(df.index))))

            # Plot
            data = {}
            if args.average:
                data['Average'] = df[['Sensor 1',
                                      'Sensor 2', 'Sensor 3']].mean(axis=1)

            elif args.sensor is not None:
                if args.sensor in [1, 2, 3, 4]:
                    col = "SENSOR {0}".format(args.sensor)
                    data[col] = df[[col]]
            else:
                for i, c in enumerate(df.columns):
                    data[c] = df[c].copy()

            if args.compare:
                if args.average:
                    data['Orig. Average'] = df_o.mean(axis=1)
                else:
                    for i in range(1, 5):
                        col = 'SENSOR %s' % i
                        new_col = 'Orig. %s' % col
                        data[new_col] = df_o[col]

            # Parse the datetime
            for k, d in data.items():
                plt.plot(d, d.index, label=k)

            plt.title(os.path.basename(f))
            plt.legend()
            plt.xlabel('Reflectance')
            plt.ylabel('Depth from Surface [cm]')
            plt.xlim([0, 4100])
            plt.show()

        except Exception as e:
            print(e)


if __name__ == '__main__':
    plot_hi_res_cli()



from radicl.high_resolution import build_high_resolution_data
import pytest
from radicl.ui_tools import get_logger
import numpy as np


@pytest.mark.parametrize('data_name, expected_data', [
    ('depth', [-100, -75, -50, -25, 0]),
    ('Y-Axis', [0, 0.5, 1, 1.5, 2.0]),
    ('Sensor1', [1000, 2000, 2500, 3000, 4000]),
])
def test_build_high_resolution_data(mock_cli, data_name, expected_data):
    log = get_logger('test_high_res')
    df = build_high_resolution_data(mock_cli, log)
    np.testing.assert_array_equal(df[data_name].values, np.array(expected_data))



import time

import pytest
from time import sleep
from . import not_connected
import numpy as np
from radicl.probe import RAD_Probe
from radicl.interface import dataframe_this


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize("setting_name, expected_type", [
    ('accrange', int)
])
def test_get_setting(real_probe, setting_name, expected_type):
    """
    Functionality test ensuring a setting runs
    """
    a = real_probe.getSetting(setting_name=setting_name)
    assert type(a) is expected_type


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize("setting_name, value", [
    ('accrange', 2),
    ('samplingrate', 5000),
    ('zpfo', 80)
])
def test_set_setting(real_probe, setting_name, value):
    """
    Test setting a parameter in the probes settings
    """
    a = real_probe.setSetting(setting_name=setting_name, value=value)
    sleep(0.1)
    a = real_probe.getSetting(setting_name=setting_name)
    assert (a == value)
    sleep(0.1)


@pytest.mark.skipif(not_connected, reason='probe not connected')
class TestProbeMeasurementData:
    @pytest.fixture(scope='class')
    def meas_probe(self):
        prb = RAD_Probe(debug=True)
        prb.setSetting(setting_name='samplingrate', value=16000)
        prb.startMeasurement()
        time.sleep(1)
        prb.stopMeasurement()
        prb.wait_for_state(3, retry=1000, delay=0.3)

        yield prb
        prb.resetMeasurement()

    # @pytest.mark.parametrize('fn, expected_n_datasets, expected_samples, ', [
    #     ('readFilteredDepthData', None, 50),
    #     ('readRawAccelerationData', 3, 75)
    #     ('readRawSensor', 3, 16000)
    #
    # ])
    # def test_dataset_len(self, meas_probe, fn, expected_samples, expected_n_datasets):
    #     """
    #     Pulls data and checks the length is as expected
    #     Args:
    #         fn: probe function attribute
    #         expected_n_datasets: number of data sets expected
    #         expected_samples: number of samples taken
    #     """
    #     data = getattr(meas_probe, fn)()
    #
    #     if expected_n_datasets is not None:
    #         keys = list(data.keys())
    #         assert len(keys) == expected_n_datasets
    #         data = data[keys[0]]
    #
    #     assert pytest.approx(len(data), abs=0.05*expected_samples) == expected_samples


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize('scale', [2, 16])
def test_rawacceleration_scaling(real_probe, scale):
    """
    If a real probe is connected, the magnitude of the sensor should
    always be very close to 1 unless the scaling is off.
    """
    a = real_probe.setSetting(setting_name='accrange', value=scale)

    real_probe.startMeasurement()
    time.sleep(0.5)
    real_probe.stopMeasurement()

    data = real_probe.readRawAccelerationData()

    data['magnitude'] = np.sqrt(np.power(data['X-Axis'], 2) +
                                np.power(data['Y-Axis'], 2) +
                                np.power(data['Z-Axis'], 2))
    result = np.mean(data['magnitude'])
    real_probe.resetMeasurement()
    assert pytest.approx(result, abs=2e-2) == 1


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize('scale', [2, 16])
def test_rawacceleration_scaling(real_probe, scale):
    """
    If a real probe is connected, the magnitude of the sensor should
    always be very close to 1 unless the scaling is off.
    """
    a = real_probe.setSetting(setting_name='accrange', value=scale)

    real_probe.startMeasurement()
    time.sleep(0.5)
    real_probe.stopMeasurement()

    data = real_probe.readRawAccelerationData()

    data['magnitude'] = np.sqrt(np.power(data['X-Axis'], 2) +
                                np.power(data['Y-Axis'], 2) +
                                np.power(data['Z-Axis'], 2))
    result = np.mean(data['magnitude'])
    real_probe.resetMeasurement()
    assert pytest.approx(result, abs=2e-2) == 1



# coding: utf-8

import datetime
import inspect
import os
import sys
import time

import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from termcolor import colored
from study_lyte.io import write_csv

from .probe import RAD_Probe
from .calibrate import get_avg_sensor
from .ui_tools import (Messages, get_logger, parse_func_list, parse_help,
                       print_helpme)

out = Messages()


def dataframe_this(data, name=None):
    """
    Takes the data returned from a function and converts it into a DataFrame
    There appears to be two scenarios.
    Args:
        data: list or dictionary of data to dataframe
        name: Name to use in the event the data is a list
    Returns:
        df: pd.Dataframe of the provided data
    """
    t = type(data)
    if t == dict:
        df = pd.DataFrame.from_dict(data)

    elif t == list:
        if isinstance(data[0], tuple):
            data = [d[0] for d in data]

        df = pd.DataFrame(data, columns=[name])

    return df


def is_numbered(filename):
    """
    Checks if the filename has been numbered. Denoted by any file ending in
    somefname_<#>.csv Returns true or false.
    Args:
        filename: Path with potential some appended number scheme denoted by _
    Returns:
        bool: True if a separator and all numbers are found towards the end
    """
    info = filename.split('.')[0]
    sep = False
    numbered = False

    if '_' in info:
        numbers = info.split('_')[-1]
        sep = True
        numbered = all([c.isnumeric() for c in numbers])
    return sep and numbered


def add_ext(filename):
    """
    Check to see if the user provided the .csv ext in the filename
    and add it
    """
    if filename[-4:] != '.csv':
        f = filename.split('.')
        # Did the user try to add an ext
        if len(f) == 2:
            filename = f[0] + '.csv'
        else:
            filename += '.csv'

    return filename


def get_default_filename():
    """
    Creates a datetime path for writing to

    Returns:
        filename: csv path named by the datetime
    """

    t = datetime.datetime.now()
    fstr = "{0}-{1:02d}-{2:02d}--{3:02d}{4:02d}{5:02d}"
    fname = fstr.format(t.year, t.month, t.day, t.hour, t.minute, t.second)
    filename = os.path.expanduser('./{0}.csv'.format(fname))

    return filename


def increment_fnumber(filename):
    """
    Check for a file numbering. Increment if there is one. Otherwise add one
    """
    no_ext = filename.split('.')[0]

    if is_numbered(filename):
        info = no_ext.split('_')
        base_file = '_'.join(info[0:-1])
        count = int(info[-1])
        fcount = count + 1

    else:
        base_file = no_ext
        fcount = 1

    filename = f"{base_file}_{fcount}.csv"

    return filename


class RADICL(object):
    """
    This is the main interface for end users to interact with the API and the
    probe.

    Attributes:
        probe: radicl
    """
    defaults = {'debug': False}

    def __init__(self, **kwargs):

        self.state = 0
        for k, v in self.defaults.items():
            if k not in kwargs.keys():
                kwargs[k] = v

        self.log = get_logger(__name__, debug=kwargs['debug'])

        self.tasks = ['daq', 'settings', 'update']
        self.task_help = {'daq': "Data acquisition using the probe.",
                          'settings': 'Interface for modifying the behavior of the probe',
                          'update': 'Firmware update dialog for the probe.'}

        self.probe = RAD_Probe(debug=kwargs['debug'])
        self.running = True

        self.settings = dir(self.probe)

        # Get all the functions available in api and probe
        probe_funcs = inspect.getmembers(
            self.probe, predicate=inspect.ismethod)

        self.options = dict()

        # Assign all data functions with keywords to auto gather data packages
        self.options['data'] = parse_func_list(probe_funcs,
                                               ['read', 'Data'],
                                               ignore_keywords=['correlation', 'integrity'])

        # Grab the settings from the probe
        self.options['settings'] = self.probe.settings
        self.options['getters'] = self.probe.getters

        # Assign a single help statement for the help option
        self.options['settings']["help"] = self.print_settings

        # User runtime preferences
        self.output_preference = None
        self.filename = None
        self.daq = None

        # Gather help dialogs
        self.help_dialog = {}

        for s in self.options.keys():

            # We haven't added the category yet
            if s not in self.help_dialog:
                self.help_dialog[s] = {}

            # Add in all the functions
            for k, fn in self.options[s].items():
                self.help_dialog[s][k] = parse_help(fn.__doc__)

    def run(self):
        """
        Main running loop for the command line Interface
        """
        # Create the probe command with an arbitrary command to have access to
        while self.running:

            pstate = self.probe.getProbeMeasState()

            if self.state == 0:
                # If probe is not ready.
                if pstate not in [0, 5]:
                    self.probe.resetMeasurement()

                else:
                    # Request type of operation
                    self.task = self.ask_user("What do you want to do with the"
                                              " probe?", self.tasks,
                                              helpme=self.task_help)

            elif self.state >= 1:
                self.tasking()

            else:
                pass

    def tasking(self):
        """
        Handles tasks from the user, specifically any member function who
        contains the keyword task_ in it. This means currently we are looking
        for a daq, settings, upgrade task function.
        """
        # Auto look for task_functions
        task_at = "task_%s" % self.task
        fn = getattr(self, task_at)

        # Call task_function
        self.log.debug("Using task function {0}".format(fn.__name__))
        fn()

    def calibrate(self):
        """
        Walks the user through calibrating
        """
        answer = self.ask_user(
            'Do you want to manually input or measure calibration values?',
            answer_lst=[
                'manual',
                'measure'])

        for i in range(1, 3):

            out.msg("For Sensor {}:".format(i))
            if answer == 'manual':
                hi = int(input('\tEnter high ADC value:\n\t'))
                lo = int(input('\n\tEnter low ADC value:\n\t'))

            elif answer == 'measure':
                input("\tApply the high value to the sensor and press enter:")
                hi = get_avg_sensor(self.probe, sensor='Sensor{}'.format(i))

                input("\n\tApply the low value to the sensor and press enter:")
                lo = get_avg_sensor(self.probe, sensor='Sensor{}'.format(i))

            else:
                raise ValueError("Invalid request for calibration.")

            self.log.info(
                "Setting calibration data to : low = {} hi = {}".format(
                    lo, hi))

            if i == 1:
                self.log.debug("Inverting for hardness...")
                low = hi
                hi = lo
                lo = low
            self.log.debug(
                "Setting calibration data to : low = {} hi = {}".format(
                    lo, hi))
            self.probe.api.MeasSetCalibData(i, lo, hi)

            values = self.probe.getSetting(setting_name='calibdata', sensor=i)
            self.log.debug("Calibdata now set to = {}".format(
                ", ".join([str(v) for v in values])))

    def take_a_reading(self):
        """
        Walks a user through the measurement taking process through pressing a
        key to start and stop measurements on the probe
        """

        input("Press any key to begin a measurement.\n")
        response = self.probe.startMeasurement()

        input("Press any key to stop the measurement.\n")
        response = self.probe.stopMeasurement()

    def listen_for_a_reading(self):
        """
        Simple CLI function to take a measurement by listening for a button
        triggering the start and stop on the probe.
        """

        out.msg(">> Press the probe button to start the measurement:")
        self.probe.wait_for_state(1, retry=1000, delay=0.3)
        out.respond("Measurement Started...")

        out.msg(">> Press the probe button to end the measurement:")
        self.probe.wait_for_state(3, retry=1000, delay=0.3)
        out.respond("Measurement ended...")

    def grab_data(self, data_request, retries=3):
        """
        Grabs data from the probe and puts it into a dataframe

        Args:
            data_request: String name of the data requesting from probe, must be a key in the self.options
            retries: Number of times to attempt to retrieve data before throwing an error
        Returns:
            data: Dataframe of the data requested
        """
        fn = self.options['data'][data_request]

        self.log.info('Downloading {} data from probe...'.format(data_request))

        attempts = 0
        success = False
        data = None

        while attempts < retries and not success:
            self.log.debug(
                "Requesting data using function {0}".format(
                    fn.__name__))

            try:
                data = fn()
                sr_ts = self.probe.getSetting(setting_name='samplingrate')
                data = dataframe_this(data, name=data_request)
                n_samples = data.index.size
                # Default to using the sensor data sample rate
                sr = sr_ts
                # Decimation ratio for peripheral sensors
                ratio = sr_ts / 16000

                if 'acceleration' in data_request:
                    sr = int(100 * ratio)
                elif data_request in ['filtereddepth', 'rawdepth', 'rawpressure']:
                    sr = int(75 * ratio)

                seconds = np.linspace(0, n_samples / sr, n_samples)
                data['time'] = seconds
                data.set_index('time', inplace=True)
                success = True

            except Exception as e:
                self.log.warning(
                    'Failed to retrieve {} data, retrying...'.format(data_request))
                self.log.debug(
                    "Failed {} attempt #{}".format(
                        data_request, attempts))
                self.log.error(e)

                success = False
                attempts += 1

        if not success:
            m = ("Unable to retrieve {} data after {} attempts"
                 "".format(data_request, attempts))
            self.log.error(m)
            data = None

        return data

    def task_daq(self):
        """
        The state machine for request to do daq
        """
        # Data Selection
        if self.state == 1:
            pstate = self.probe.getProbeMeasState()
            self.log.info("Checking to see if probe is ready...")
            self.log.debug("Probe State = {0}".format(pstate))

            # Make sure probe is ready
            if pstate == 0 or pstate == 5:
                self.daq = self.ask_user("What data do you want?",
                                         list(self.options['data'].keys()),
                                         helpme=self.help_dialog['data'])

        # Method for outputting data
        elif self.state == 2:
            self.output_preference = self.ask_user("How do you want to output"
                                                   " {0} data?".format(
                self.daq),
                ['plot', 'write', 'both'],
                default_answer=self.output_preference)

        # Take Measurements
        elif self.state == 3:
            self.take_a_reading()

            self.data = self.grab_data(self.daq)
            # acc_cols = [c for c in self.data.columns if 'Axis' in c]
            # self.data[acc_cols] = self.data[acc_cols].mul(2)

            self.state = 4

            if self.data is None:
                self.state = 3
                self.log.error('')
                self.log.error('Retrieving probe data failed, try again.')

            response = self.probe.resetMeasurement()

        # Data output and options
        elif self.state == 4:
            if self.output_preference in ['write', 'both']:
                valid = False

                #  Wait for real path
                while not valid:
                    fname = get_default_filename()
                    out.msg("Enter in a filepath for the data to be saved:")
                    filename = input("\nPress enter to use default:"
                                     "\n(Default: {0})".format(fname))

                    # Assign a default path
                    if filename == '':
                        filename = fname

                    # Make it absolute
                    filename = os.path.abspath(filename)
                    real_path = os.path.isdir(os.path.dirname(filename))

                    # Double check a real path was given
                    if not real_path:
                        out.warn("Path provided does not exist.")
                        valid = False
                    else:
                        self.filename = filename
                        valid = True
                        self.log.info(
                            "Saving Data to :\n{0}".format(
                                self.filename))

                        if not self.data.empty:
                            self.write_probe_data(self.data, filename=filename)
                            self.state = 5

            if self.output_preference in ['plot', 'both']:
                self.data.plot()
                plt.show()
                self.state = 5

        elif self.state == 5:
            response = self.ask_user("Take another measurement?")

            if response in ['y', 'yes', True]:
                self.state = 3

            # Go to DAQ menu
            else:
                self.state = 0

    def task_settings(self):
        """
        Routine for handling users requests for modifying probe settings settings
        """

        # Setting Selection
        if self.state == 1:
            pstate = self.probe.getProbeMeasState()
            self.log.info("Checking to see if probe is ready...")
            self.log.debug("Probe State = {0}".format(pstate))

            # Make sure probe is ready
            if pstate in [0, 5]:
                self.setting_request = \
                    self.ask_user("What setting do you want to adjust?",
                                  sorted(list(self.probe.settings.keys())),
                                  helpme=self.help_dialog['settings'])

        # Get current setting
        elif self.state == 2:

            if self.setting_request == 'show':
                self.options['settings'][self.setting_request]()
                self.state = 1

            # Retrieve calibration values
            elif self.setting_request == 'calibdata':
                for i in range(1, 3):
                    self.current_setting_value = []
                    self.current_setting_value.append(
                        self.probe.getSetting(setting_name=self.setting_request,
                                              sensor=i))
                    values = ", ".join([str(v)
                                        for v in self.current_setting_value])
                    msg = ("Currently {0}[{2}] = {1}\nEnter value to change probe {0}\n"
                           "".format(self.setting_request,
                                     values, i))
                    self.log.info(msg)
                self.state += 1

            else:
                self.current_setting_value = \
                    self.probe.getSetting(setting_name=self.setting_request)
                self.state += 1

        # Modify setting
        elif self.state == 3:

            if self.setting_request == 'calibdata':
                self.calibrate()
                self.state = 1

            else:
                values = self.current_setting_value
                msg = ("Currently {0} = {1}\nEnter value to change probe {0}\n"
                       "".format(self.setting_request,
                                 values))
                self.log.info(msg)

                valid = False

                # All entries have to be numeric. Ensure this is the case.
                while not valid:
                    self.new_value = input(msg)
                    try:
                        self.new_value = int(self.new_value)
                        valid = True
                    except Exception as e:
                        out.error(e)
                        out.error("Value must be numeric!")

                # Call the function to change the setting
                self.probe.setSetting(setting_name=self.setting_request,
                                      value=self.new_value)
                time.sleep(0.2)

                # Confirm the value was changed
                test_value = self.probe.getSetting(
                    setting_name=self.setting_request)

                if test_value == self.new_value:
                    self.log.info("{0} was changed from {1} to {2}!\n"
                                  "".format(self.setting_request,
                                            self.current_setting_value,
                                            self.new_value))
                time.sleep(1)

                # Go back to settings menu
                self.state = 1

    def print_settings(self):
        """
        print all the current settings

        helpme - Prints out all the settings for the probe

        """

        msg = "\n===== Current Probe Settings =====\n"

        for s, fn in self.probe.getters.items():
            if s.lower() not in ['calibdata', 'numsegments']:
                value = self.probe.getSetting(setting_name=s)
                msg += "{} = {}\n".format(s, value)
        out.msg(msg)
        time.sleep(2)

    def write_probe_data(self, df, filename='', extra_meta={}):
        """
        Writes out a dataframe with a probe header to csv
        Args:
            df: pandas dataframe containing data
            filename: valid path to output to, if empty uses datetime
            extra_meta: Dictionary of extra notes to add to the file header
        """

        # Receive a default request
        if filename == '':
            filename = get_default_filename()

        else:
            filename = os.path.expanduser(filename)

        out.msg("Saving Data to :\n{0}".format(filename))

        if not df.empty:
            # Write the header so we know things about this
            meta = self.probe.getProbeHeader()
            meta.update(extra_meta)
            write_csv(df, meta, filename)

    def ask_user(self, question_str, answer_lst=None, helpme=None,
                 next_state=True, default_answer=None):
        """
        Function is used to wait for an appropriate response from user and handle unknown answers

        Args:
            question_str - String of the question
            answer_lst - list of acceptable answers
            helpme - help strings in a dictionary to elaborate on acceptable answers
            next_state - Advance the state of the cli state machine
            default_answer - if the user selects nothing and hits enter the answer will default to this

        Returns:
            response - string of the valid response from the user
        """

        # Ask user for a yes no question
        if answer_lst is None:
            answer_lst = ['y', 'n']
            question_boolean = True
        else:
            question_boolean = False

        # Keep question to one line if small number of options
        lower_lst = [i.lower() for i in sorted(answer_lst)]

        # Add in option to return to main menu
        if self.state != 0:
            lower_lst.append('home')

        # Help documentation
        if helpme is not None:
            lower_lst.append("help")

        # Exit the program
        lower_lst.append('exit')

        # If the options list if small keep it on the same line
        if len(lower_lst) <= 5:
            print_able = " ("
            print_able += ", ".join(lower_lst)
            print_able += ")"

        # If the list of choices is long create a column of choices
        else:
            print_able = '\n\n{0}'.format(colored('[OPTIONS]', 'magenta',
                                                  attrs=['bold']))
            for s in lower_lst:
                print_able += '\n  {0}'.format(s)

        question_str += print_able

        # Provide prompt for default options
        if default_answer is not None:
            if default_answer not in answer_lst:
                raise ValueError("Default answer must be a member of the "
                                 "acceptable answers list in ask_user function")

            question_str += '(Default: {0})'.format(default_answer)

        out.msg(question_str)

        acceptable_answer = False

        while not acceptable_answer:
            user_answer = input('\n')
            response = user_answer.lower().strip()

            # User requests help with options
            if "help" in response:
                print_helpme(response, helpme)
                acceptable_answer = False
                out.msg(question_str)

            # Request to leave the program
            elif response == 'exit':
                self.log.info("Exiting the RAD CLI.")
                self.running = False
                sys.exit()

            elif response == 'home':
                self.log.info("Returning to the main menu.")
                self.running = True
                acceptable_answer = True
                self.state = 0

            # Default Picked
            elif response == '':
                response = default_answer

            # User provided a potential acceptable answer
            elif response in lower_lst:
                acceptable_answer = True
                response = lower_lst[lower_lst.index(response)]
                if question_boolean:
                    if response == 'y' or response == 'yes':
                        response = True
                    else:
                        response = False

                if next_state:
                    self.state += 1

            else:
                out.error('Only acceptable answers are:')
                out.msg(print_able)
                acceptable_answer = False

        return response



import pytest
from . import MockRADPort
from radicl.api import RAD_API


class TestApi:
    @pytest.fixture()
    def mock_port(self, payload):
        yield MockRADPort(payload)

    @pytest.fixture()
    def mock_api(self, mock_port):
        return RAD_API(mock_port, debug=True)

    @pytest.mark.parametrize('payload, expected', [
        [b'\x9f\x01\x02\x00\x01\x03', 3]
    ])
    def test_getHWID(self, mock_api, payload, expected):
        ret = mock_api.getHWID()
        assert ret['data'] == expected

    @pytest.mark.parametrize('payload, expected', [
        [b'\x9f\x02\x02\x00\x01\x01', 1]
    ])
    def test_getHWREV(self, mock_api, payload, expected):
        ret = mock_api.getHWREV()
        assert ret['data'] == expected

    @pytest.mark.parametrize('payload, expected', [
        [b'\x9f\x03\x02\x00\x02\x01.', 1.46]
    ])
    def test_getFWREV(self, mock_api, payload, expected):
        ret = mock_api.getFWREV()
        assert ret['data'] == expected

    @pytest.mark.parametrize('payload, expected', [
        [b'\x9f\t\x02\x00\x04\x01.\x03\x00', '1.46.3.0']
    ])
    def test_getFullFWREV(self, mock_api, payload, expected):
        ret = mock_api.getFullFWREV()
        assert ret['data'] == expected



import pytest

from . import MockGPSStream
from radicl.gps import USBGPS
from unittest.mock import patch
from types import SimpleNamespace


@pytest.fixture(scope='function')
def mock_gps_port(payload):
    yield MockGPSStream(payload)


@pytest.mark.parametrize('payload, expected', [
    # Lat long found with a retry
    ([b'$GPTXT,01,01,02,u-blox ag - www.u-blox.com*50\r\n',
      b'$GPRMC,201209.00,A,4400.0000,N,11600.00000,W,0.065,,230223,,,D*6B\r\n'], [44.0, -116.0]),
    # No lat long found
    ([b'$GPTXT,01,01,02,u-blox ag - www.u-blox.com*50'], None),
    # Lat long message found but not interpretable
    ([b'$GPRMC,201209.00,A,,N,,W,0.065,,230223,,,D*6B\r\n'], None)
])
def test_get_gps_fix(mock_gps_port, payload, expected):
    """
    Mock out the connection and gps. Ensure the managing function handles
    three scenarios.
    Args:
        payload: List of nmea strings to read
        expected: expected lat long outcome
    """
    with patch('radicl.gps.get_serial_cnx', return_value=SimpleNamespace(port='dev_fake', description='GPS/GNSS Receiver')):
        with patch('radicl.com.serial.Serial.open', return_value=None):
            with patch('radicl.gps.NMEAReader', return_value=mock_gps_port):
                gps_dev = USBGPS()
                loc = gps_dev.get_fix(max_attempts=2)
                assert loc == expected



# coding: utf-8

import time

from .ui_tools import get_logger
from .commands import MeasCMD, SystemCMD, SettingsCMD, FWUpdateCMD, AttributeCMD

PCA_ID_LIST = ["UNKNOWN", "PB1", "PB2", "PB3"]


class RAD_API:
    """
    Class for directly interacting with the probe in a non-human friendly way
    """

    def __init__(self, port, debug=False):
        self.port = port
        self.log = get_logger(__name__, debug=debug)
        self._serial = None
        self._hw_id = None
        self._hw_id_str = None
        self._hw_rev = None
        self._fw_rev = None
        self._full_fw_rev = None

    def __sendCommand(self, data):
        """
        Generic send function
        Returns 1 of successful, 0 otherwise
        """

        success = 0
        try:
            # Send the data
            self.port.writePort(data)

        except Exception as e:
            self.log.error(e)

        else:
            success = 1

        finally:
            return success

    def __getResponse(self):
        """
        Generic read function
        Returns 1 if successful, 0 otherwise
        """

        success = 0
        response = ""

        try:
            num_bytes_in_buffer = self.port.numBytesInBuffer()
            if num_bytes_in_buffer > 0:
                response = self.port.readPort(num_bytes_in_buffer)

        except Exception as e:
            self.log.error(e)

        else:
            success = 1

        finally:
            if success:
                return response
            else:
                return None

    def __send_receive(self, data, read_delay=0.05):
        """
        Generic send/receive function
        Returns the response if successful, empty result otherwise
        """
        # Dynamic delay variable, increases with each failed loop
        delay_counter = 0.001

        # Make sure that the read delay is at least 1ms (i.e. it cannot be 0)
        if read_delay < 0.001:
            read_delay = 0.001

        ret = self.__sendCommand(data)

        if ret:
            # The command was successfuly sent. Now read the response
            time.sleep(0.001)
            ret = self.__getResponse()
            # Read the response as long as we don't get anything back AND
            # we haven't exceeded the max. read delay
            while (((ret is None) or (ret == ""))
                   and (delay_counter < read_delay)):
                time.sleep(delay_counter)
                delay_counter += 0.001
                ret = self.__getResponse()
            # If we get here either we have exceeded the max. read delay, or
            #  we have received a response.
            return ret

        else:
            # There was an issue with sending the command
            return None

    def __isACK(self, message, cmd=None):
        """
        Returns 1 if the message contains an ACK
        """
        if len(message) >= 5:
            # If a command was specified, check that it matches the message
            if cmd is not None:
                if message[1] != cmd:
                    return 0
            # If we get here then we have passed the additional specified
            # checks.
            if message[2] == 0x04:
                # ACK detectec
                return 1
            else:
                # No ACK found
                return 0
        else:
            # Message is not long enough. Indicate an error
            return 0

    def __isNACK(self, message, cmd=None):
        """
        Returns 1 if the message contains a NACK
        """

        if len(message) >= 5:

            # If a command was specified, check that it matches the message
            if cmd is not None:
                if message[1] != cmd:
                    return 0

            # If we get here then we have passed the additional user specified
            # checks.
            if message[2] == 0x05:
                # NACK detectec
                return 1
            else:
                # No NACK found
                return 0
        else:
            # Message is not long enough. Indicate an error
            return 0

    def __getNACKValue(self, message):
        """
        Returns the NACK message value
        If an error is detected, this will simply return 0
        """

        if len(message) >= 7:
            nack_val = message[5:7]
            return int.from_bytes(nack_val, byteorder='little')

        else:
            return 0

    def __isResponse(self, message, cmd=None, data_len=None):
        """
        Returns 1 if the message is a valid response
        0 for data_len means that the payload length could be variable
        """
        valid = 0
        length = len(message)
        if length >= 5:
            if message[2] == 0x02:
                calc_len = message[4] + 5
                # If a command was specified we need to check if it matches.
                # If it is incorrect, the response is invalid
                if message[1] == cmd:
                    valid = 1
                # If a data length was specified we need to check if it
                # matches the message's data payload length. If it is incorrect
                # we will return immediately
                if (data_len is not None) and (data_len != 0):
                    if message[4] == data_len:
                        valid = valid * 1

                # If we get here then we have passed the additional
                # specified checks. Finally, check if the overall length
                # matches (integrity check)
                if length == calc_len:
                    # The length is correct
                    valid = valid * 1
                else:
                    # Not a response or incorrect length
                    valid = 0

            elif (message[2] == 0x06) and (length == 261):
                valid = 1

        return valid

    def __getNumPayloadBytes(self, message):
        """
        Returns the number of bytes present in the payload. This takes long #
        responses into account
        """

        if message[2] == 0x02 or message[2] == 0x03:
            return message[4]
        elif message[2] == 0x06:
            return 256
        else:
            return 0

    def __isPushMessage(self, message, cmd=None, data_len=None):
        """
        Returns 1 if the message is a valid push message/response
        """
        length = len(message)
        if length >= 5:
            calc_len = message[4] + 5
            # If a command was specified we need to check if it matches.
            # If it is incorrect we will return immediately
            if cmd is not None:
                if message[1] != cmd:
                    return 0
            # If a data length was specified we need to check if it matches the
            #  message's data payload length. If it is incorrect we will
            # return immediately
            if data_len is not None:
                if message[4] != data_len:
                    return 0
            # If we get here then we have passed the additional user specified
            # checks. Finally, check if the message is a response and if the
            # overall length matches (integrity check)
            if (message[2] == 0x03) and (length == calc_len):
                # This is a response and the length is correct
                return 1
            else:
                # Not a response or incorrect length
                return 0
        else:
            # Message is not long enough. Indicate an error
            return 0

    def __waitForMessage(self, timeout, expected_bytes=5):
        """
        This function simply waits for a message. It is like a read, but waits
        up to the specified timeout for new data to arrive
        Args:
                timeout: specified in seconds. The smallest delay period is 0.01s
        """

        delay_time = 0.01
        # Force the timeout to be at least 10ms
        if timeout < delay_time:
            timeout = delay_time
        num_iter = timeout / delay_time
        response = []
        num_bytes_in_buffer = 0
        while num_iter:
            try:
                num_bytes_in_buffer += self.port.numBytesInBuffer()

                if num_bytes_in_buffer > 0:
                    response.extend(self.port.readPort(num_bytes_in_buffer))
            except Exception as e:
                self.log.error(e)
            else:
                if num_bytes_in_buffer >= expected_bytes:
                    return response
                else:
                    num_iter = num_iter - 1
                    time.sleep(delay_time)
        return None

    def __EvaluateAndReturn(self, response, expected_command,
                            num_expected_payload_bytes):
        """
        This function evaluates the response and prepares the API return value
        If 'num_expected_payload_bytes' is 0, then the response guides how
        many bytes will be returned

        Args:
                response: data in bytes from the probe received from __send_receive
                expected_command:
                num_expected_payload_bytes: Integer number of bytes expecting to receive
        """

        if response is None:
            result = {'status': 0, 'errorCode': None, 'data': None}

        elif (self.__isResponse(response, expected_command,
                                num_expected_payload_bytes)):
            result = {'status': 1, 'errorCode': None,
                      'data': response[-(
                          self.__getNumPayloadBytes(response)):]}

        elif (self.__isPushMessage(response, expected_command,
                                   num_expected_payload_bytes)):
            result = {'status': 1, 'errorCode': None,
                      'data': response[-(
                          self.__getNumPayloadBytes(response)):]}

        elif self.__isACK(response):
            result = {'status': 1, 'errorCode': None, 'data': None}

        elif self.__isNACK(response):
            nack_value = self.__getNACKValue(response)
            result = {'status': 0, 'errorCode': nack_value, 'data': None}

        else:
            result = {'status': 0, 'errorCode': None, 'data': None}
        return result

    # ********************
    # * PUBLIC FUNCTIONS *
    # ********************

    def sendApiPortEnable(self):
        """
        Sends a '*' to enable the API port which tells the probe to interact
        via the radicl API
        """
        self.port.writePort([0x21])

    @property
    def hw_id(self):
        """
        Gets the hardware id
        """
        if self._hw_id is None:
            ret = self.getHWID()
            self._hw_id = ret['data']
        return self._hw_id

    @property
    def hw_id_str(self):
        """
        Gets the hardware id string
        """
        if self._hw_id_str is None:
            if self.hw_id < len(PCA_ID_LIST):
                self._hw_id_str = PCA_ID_LIST[self.hw_id]
        return self._hw_id_str

    @property
    def hw_rev(self):
        """
        Gets the hardware revision
        """
        if self._hw_rev is None:
            ret = self.getHWREV()
            self._hw_rev = ret['data']
        return self._hw_rev

    @property
    def fw_rev(self):
        """
        Gets the hardware id
        """
        if self._fw_rev is None:
            ret = self.getFWREV()
            self._fw_rev = ret['data']
        return self._fw_rev

    @property
    def full_fw_rev(self):
        """
        Gets the hardware id
        """
        if self._full_fw_rev is None:
            ret = self.getFullFWREV()
            self._full_fw_rev = ret['data']
        return self._full_fw_rev

    @property
    def serial(self):
        """
        Gets the serial string
        """
        if self._serial is None:
            self._serial = self.getSerialNumber()
        return self._serial

    def Identify(self):
        """
        Identifies the connected device
        Returns 1 if a valid device was detected
        """
        self.log.debug("Retrieving probe information...")

        if self.hw_id is not None:
            if self.hw_id < len(PCA_ID_LIST):
                id_msg = f"Attached device: {self.hw_id_str}, " \
                         f"Revision={self.hw_rev}, "

                if self.full_fw_rev is not None:
                    id_msg += f"Firmware = {self.full_fw_rev}"
                else:
                    id_msg += f"Firmware = {self.fw_rev}"

                self.log.info(id_msg)
                return 1

            else:
                self.log.warning("Unknown device detected!")
                return 0

        else:
            self.log.warning("Invalid response to ID request")
            return 0

    # ***************************
    # ***** BASIC COMMANDS ******
    # ***************************
    def getSerialNumber(self):
        """
        Queries the board's serial number
        """
        code = AttributeCMD.SERIAL.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 8)

    def getHWID(self):
        """
        Queries the board's HW ID
        """
        code = AttributeCMD.HW_ID.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        ret_val = self.__EvaluateAndReturn(response, code, 1)

        if ret_val['status'] == 1:
            data = int.from_bytes(ret_val['data'], byteorder='little')
            ret_val['data'] = data

        return ret_val

    def getHWREV(self):
        """
        Queries the board's HW revision
        """
        code = AttributeCMD.HW_REV.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        ret_val = self.__EvaluateAndReturn(response, code, 1)
        if ret_val['status'] == 1:
            byte_arr = ret_val['data']
            data = int.from_bytes(ret_val['data'], byteorder='little')
            ret_val['data'] = data

        return ret_val

    def getFWREV(self):
        """
        Queries the board's FW revision in MAJOR.MINOR format
        """
        code = AttributeCMD.FW_REV.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        ret_val = self.__EvaluateAndReturn(response, code, 2)
        if ret_val['status'] == 1:
            major = ret_val['data'][0]  # value[-2]
            minor = ret_val['data'][1]  # value[-1]
            fw_rev = major + minor / 100
            ret_val['data'] = fw_rev
        return ret_val

    def getFullFWREV(self):
        """
        Queries the board's FW revision in the full A.B.C.D format
        """
        code = AttributeCMD.FULL_FW_REV.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        ret_val = self.__EvaluateAndReturn(response, code, 4)
        if ret_val['status'] == 1:
            value = ret_val['data']
            rev_string = str(value[0]) + "." + str(value[1]) + \
                         "." + str(value[2]) + "." + str(value[3])
            ret_val['data'] = rev_string
        return ret_val

    def startBootloader(self):
        """
        Starts the bootloader
        """
        code = SystemCMD.START_BOOTLOADER.cmd
        response = self.__send_receive([0x9F, code, 0x01, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)

    def getSystemStatus(self):
        """
        Queries the system status
        """
        code = SettingsCMD.STATUS.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 4)

    def getRunState(self):
        code = SystemCMD.STATE.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    # ***************************************
    # ***** SENSOR/MEASUREMENT COMMANDS *****
    # ***************************************

    def getMeasState(self):
        """
        Queries the state of the measurement state machine
        """
        code = MeasCMD.STATE.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasReset(self):
        """
        Resets the measurement state machine
        Returns 1 if successful
        """
        code = MeasCMD.RESET.cmd
        response = self.__send_receive([0x9F, code, 0x01, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasStart(self):
        """
        Starts a measurement
        Returns 1 if successful
        """
        code = MeasCMD.START.cmd
        response = self.__send_receive([0x9F, code, 0x01, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasStop(self):
        """
        Stops a measurement
        Returns 1 if successful
        """
        code = MeasCMD.STOP.cmd
        response = self.__send_receive([0x9F, code, 0x01, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetNumSegments(self, buffer_id):
        """
        Queries the number of data segments for a particular data buffer
        """
        code = MeasCMD.NUM_SEGMENTS.cmd
        message = [0x9F, code, 0x00, 0x00, 0x01]
        message.extend(buffer_id.to_bytes(1, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasReadDataSegment(self, buffer_id, numPacket):
        """
        Reads a specific data segment of a specific data buffer
        """
        code = MeasCMD.DATA_SEGMENT.cmd
        message = [0x9F, code, 0x00, 0x00, 0x05]
        message.extend(buffer_id.to_bytes(1, byteorder='little'))
        message.extend(numPacket.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)

        # Check if only the command matches. The length may be variable
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetSamplingRate(self):
        """
        Reads/Returns the IR sampling rate
        """
        code = SettingsCMD.SAMPLING_RATE.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasSetSamplingRate(self, sampling_rate):
        """
        Returns status=1 if successful, status=0 otherwise

        helpme - Sets the sensor sampling rate
        """
        code = SettingsCMD.SAMPLING_RATE.cmd
        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(sampling_rate.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetZPFO(self):
        """
        Reads/returns the Zero Phase Filter Order used on the depth data.
        """
        code = SettingsCMD.ZPFO.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasSetZPFO(self, zpfo):
        """
        Returns status=1 if successful, status=0 otherwise

        helpme - Set the Zero Phase Filter Order used on the depth data.
        """
        code = SettingsCMD.ZPFO.cmd
        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(zpfo.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetPPMM(self):
        """
        Reads/returns the Points per millimeter parameter
        """
        code = SettingsCMD.PPMM.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasSetPPMM(self, ppmm):
        """
        helpme - Sets the Points per millimeter parameter
        """
        code = SettingsCMD.PPMM.cmd
        message = [0x9F, code, 0x01, 0x00, 0x01]
        message.extend(ppmm.to_bytes(1, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetALG(self):
        """
        Reads/Returns the algorithm (1 - depth corrected, 2 for timeseries only)
        parameter
        """
        code = SettingsCMD.ALG.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasSetALG(self, alg):
        """
        Returns status=1 if successful, status=0 otherwise

        helpme - Sets the algorithm (1 - depth corrected, 2 for timeseries only)
        """
        code = SettingsCMD.ALG.cmd
        message = [0x9F, code, 0x01, 0x00, 0x01]
        message.extend(alg.to_bytes(1, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetAPPP(self):
        """
        Reads the APPP parameter
        """
        code = SettingsCMD.APPP.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasSetAPPP(self, appp):
        """
        Returns status=1 if successful, status=0 otherwise
        helpme - Sets the APPP parameter which smooths the timeseries data
        """
        code = SettingsCMD.APPP.cmd
        message = [0x9F, code, 0x01, 0x00, 0x01]
        message.extend(appp.to_bytes(1, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetTCM(self):
        """
        Reads the TCM parameter
        """
        code = SettingsCMD.TCM.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasSetTCM(self, tcm):
        """
        Returns status=1 if successful, status=0 otherwise

        helpme - Sets the Temperature correction method for the barometer data

        """
        code = SettingsCMD.TCM.cmd

        message = [0x9F, code, 0x01, 0x00, 0x01]
        message.extend(tcm.to_bytes(1, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetUserTemp(self):
        """
        Reads the user set temperature
        """
        code = SettingsCMD.USERTEMP.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasSetUserTemp(self, user_temp):
        """
        Returns status=1 if successful, status=0 otherwise

        helpme - Set the user specified temperature for TCM=3
        """
        code = SettingsCMD.USERTEMP.cmd

        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(user_temp.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetIR(self):
        """
        Reads the IR parameter

        """
        code = SettingsCMD.IR.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasSetIR(self, ir):
        """
        Returns status=1 if successful, status=0 otherwise

        helpme - Turns on the IR emitter
        """
        code = SettingsCMD.IR.cmd
        message = [0x9F, code, 0x01, 0x00, 0x01]
        message.extend(ir.to_bytes(1, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetCalibData(self, num_sensor):
        """
        Reads a sensor's calibration value
        """
        code = SettingsCMD.CALIBDATA.cmd
        message = [0x9F, code, 0x00, 0x00, 0x01]
        message.extend(num_sensor.to_bytes(1, byteorder='little'))

        response = self.__send_receive(message)

        # Expect 4 bytes back two for the low value and 2 for the high value
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasSetCalibData(self, num_sensor, calibration_value_low,
                         calibration_value_high):
        """
        Sets the probes calibration values. A high and low are set where the
        low. This is applied linearly and thus the low value should be the
        y intercept. This will remap the raw's low value-high value to 0-4095

        The probe expects this number joined in bytes so we have to create a
        2, 2 byte integers and add them together to create a 4 byte message.

        Args:
                num_sensor: Integer indicating sensors 1,2,3, or 4.
                calibration_value_low: 12-bit integer indicating the y intercept of a linear
                                   calibration
                calibration_value_high: 12-bit integer for the

        Returns:
                status: 1 if successful, 0 otherwise

        helpme - Sets the calibration data for the specified sensor
        """
        code = SettingsCMD.CALIBDATA.cmd
        message = [0x9F, code, 0x01, 0x00, 0x05]

        # Convert values each into a 1 and 2 bytes
        message.extend(num_sensor.to_bytes(1, byteorder='little'))
        message.extend(calibration_value_low.to_bytes(2, byteorder='little'))
        message.extend(calibration_value_high.to_bytes(2, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetMeasTemp(self):
        """
        Reads the current temperature reading (from last measurement)
        Returns status=1 if successful, status=0 otherwise
        """
        code = MeasCMD.TEMP.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetAccThreshold(self):
        """
        Reads the accelerometer threshold setting (an unsigned 32-bit integer in mG)
        Returns status=1 if successful, status=0 otherwise
        """
        code = SettingsCMD.ACCTHRESH.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasSetAccThreshold(self, threshold):
        """
        Sets the accelerometer threshold setting (accelerometer thresholding algorithm)
        The parameter 'threshold' is an unsigned 32-bit (4-bytes) absolute value indicating the
        threshold in mG A value of 0 turns the accelerometer thresholding algorithm off
        Returns status=1 if successful, status=0 otherwise
        """
        code = SettingsCMD.ACCTHRESH.cmd
        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(threshold.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetAccZPFO(self):
        """
        Reads the accelerometer zero-phase filter order setting (post-processing filter
        for accelerometer thresholding algorithm) Returns status=1 if successful,
        status=0 otherwise
        """
        code = SettingsCMD.ACCZPFO.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 4)

    def MeasSetAccZPFO(self, zpfo):
        """
        Sets the accelerometer zero-phase filter order setting (post-processing filter
        for accelerometer thresholding algorithm) The parameter 'zpfo' is an
        unsigned 32-bit (4-bytes) value indicaing the filter order A value of 0
        turns the filtering off (filter is bypassed) Returns status=1 if successful,
        status=0 otherwise
        """
        code = SettingsCMD.ACCZPFO.cmd
        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(zpfo.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def MeasGetAccRange(self):
        """
        gets the accelerometer range
        """
        code = SettingsCMD.ACCRANGE.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def MeasSetAccRange(self, abs_range_gs):
        """
        Sets the accelerometer range

        Args:
            abs_range_gs: integer value for range e.g. 2 == +/-2g's

        Returns:
        """
        code = SettingsCMD.ACCRANGE.cmd
        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(abs_range_gs.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    # ******************************
    # ***** FW UPDATE COMMANDS *****
    # ******************************

    def UpdateEnter(self):
        """
        Enters the FW Update FSM
        Returns status=1 if successful, status=0 otherwise
        """
        code = FWUpdateCMD.ENTER.cmd
        response = self.__send_receive([0x9F, code, 0x01, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)

    def UpdateGetState(self):
        """
        Gets the FW update FSM state
        """
        code = FWUpdateCMD.STATE.cmd
        response = self.__send_receive([0x9F, code, 0x00, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 1)

    def UpdateWaitForStateChange(self, wait_time):
        """
        Waits for the state change message
        """
        code = FWUpdateCMD.STATE.cmd
        response = self.__waitForMessage(wait_time, 6)
        return self.__EvaluateAndReturn(response, code, 1)

    def UpdateSetSize(self, num_packets, packet_size):
        """
        Sets the update size
        Returns status=1 if successful, status=0 otherwise
        """
        code = FWUpdateCMD.SIZE.cmd
        message = [0x9F, code, 0x01, 0x00, 0x06]
        message.extend(num_packets.to_bytes(4, byteorder='little'))
        message.extend(packet_size.to_bytes(2, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def UpdateDownload_Short(self, data, crc8, packet_id):
        """
        Downloads a 16-byte chunk of data
        Returns status=1 if successful, status=0 otherwise
        """
        code = FWUpdateCMD.DOWNLOAD.cmd
        message = [0x9F, code, 0x01, 0x00, 0x00]
        message[3] = crc8
        message[4] = len(data)
        message.extend(data)
        self.__sendCommand(message)
        response = self.__waitForMessage(20)
        return self.__EvaluateAndReturn(response, code, 0)

    def UpdateDownload_Long(self, data, crc8, packet_id):
        """
        Downloads a 256-byte chunck of data
        Returns status=1 if successful, status=0 otherwise
        """
        code = FWUpdateCMD.DOWNLOAD.cmd
        message = [0x9F, code, 0x07, 0x00, 0x00]
        message[3] = crc8
        message[4] = 0
        message.extend(data)
        self.__sendCommand(message)
        response = self.__waitForMessage(20)
        return self.__EvaluateAndReturn(response, code, 0)

    def UpdateSetCRC(self, crc32):
        """
        Sets the CRC32 of the FW image
        Returns status=1 if successful, status=0 otherwise
        """
        code = FWUpdateCMD.CRC.cmd
        message = [0x9F, code, 0x01, 0x00, 0x04]
        message.extend(crc32.to_bytes(4, byteorder='little'))
        response = self.__send_receive(message)
        return self.__EvaluateAndReturn(response, code, 0)

    def UpdateClose(self):
        """
        Closes the FW update FSM
        Returns status=1 if successful, status=0 otherwise
        """
        code = FWUpdateCMD.CLOSE.cmd
        response = self.__send_receive([0x9F, code, 0x01, 0x00, 0x00])
        return self.__EvaluateAndReturn(response, code, 0)



# coding: utf-8

import os
import time
import numpy as np

from radicl.ui_tools import get_logger


def get_avg_sensor(probe, delay=3, sensor='Sensor1'):
    """
    Take a probe measurement of 3 seconds and return the mean of that data.

    Args:
        probe: radicl.probe.RAD_Probe object
        delay: Measurement time in seconds
        sensor: String name of the column of the data to avg
    Returns:
        probe_data: Mean of the Raw Sensor over the delay time period
    """
    log = get_logger(__name__)

    # Start the probe
    probe.startMeasurement()
    log.info("\tProbe measurement started...")
    time.sleep(delay)

    # Stop probe
    probe.stopMeasurement()
    log.info("Measurement Stopped.")

    # Probe data extraction and average
    probe_data = np.array((probe.readRawSensorData())[sensor])
    probe_data = probe_data[probe_data < 4096].mean()

    # Start the reset in another thread
    probe.resetMeasurement()

    # If we like the data
    if probe_data in [None, 65535]:
        log.error("Probe data invalid! Trying again")

    return int(probe_data)


if __name__ == '__main__':
    main()



# coding: utf-8

import serial
from serial.tools import list_ports
from .ui_tools import get_logger


def find_kw_port(kw):
    """
    Find a com port by looking at Keyword attributes of the port

    Args:
            kw: List of kw to use (not case sensitive)
    Return:
            match_list: list of port names that have at least one matching keyword
    """
    match_list = []

    # Run through all available COM ports grabs ones that match our keywords
    port_list = list_ports.comports()

    for p in port_list:
        # Make a list of true for every keyword we find in the port data
        kw_match = [k.lower() in p.description.lower() for k in kw]

        # If the match list is not empty append this port name
        if any(kw_match):
            match_list.append(p)

    return match_list


def get_serial_cnx(keyword, match_index=0):
    """
    Use a keyword to return any matches in the serial com ports.
    Then return the match of interest using the match_index kwarg.

    Args:
        keyword: String keyword to search for in device descriptions
        match_index: Index to use if more than one match comes back

    Returns:
        cnx: serial.Serial object
    """
    cnx = None
    matching_ports = find_kw_port([keyword])
    if match_index < len(matching_ports):
        port = matching_ports[match_index]
        cnx = serial.Serial(port.device)
    return cnx


class RAD_Serial:
    """
    This class handles all serial communication and simply acts
    as an abstraction layer
    """

    def __init__(self, debug=False):
        self.serial_port = None
        self.log = get_logger(__name__, debug=debug)

    def openPort(self, com_port=None):

        # No COM port has been provided. Need to detect port automatically
        if com_port is None:
            self.log.info("No COM port provided. Scanning for COM ports...")
            match_list = find_kw_port(['STMicroelectronics', 'STM32'])

            # Check if more than one was found
            if len(match_list) > 1:
                self.log.warn('Multiple COM ports found, using the first')

            # If no ports match then error out gracefully
            elif not match_list:
                self.log.error("No serial ports were found for the Lyte probe!")
                raise serial.SerialException('No comports were found!')

            # Finally, assign the found port to the serial_port variable
            com_port = match_list[0].device

        try:
            self.log.debug(
                "Attempting to establish a connection with the probe...")
            self.serial_port = serial.Serial(port=com_port,
                                             baudrate=115200,
                                             parity=serial.PARITY_NONE,
                                             stopbits=serial.STOPBITS_ONE,
                                             bytesize=serial.EIGHTBITS,
                                             timeout=0.01,
                                             write_timeout=0,
                                             xonxoff=False,
                                             rtscts=False,
                                             dsrdtr=False)
            self.serial_port.setDTR(1)
            self.log.info("Using {}".format(self.serial_port.port))

        except Exception as e:
            self.log.error("Serial port open failed: {}".format(e))
            raise IOError("Could not open COM port")

    def closePort(self):
        if self.serial_port is not None:
            self.serial_port.close()
            self.serial_port = None

    def flushPort(self):
        if self.serial_port is not None:
            self.serial_port.reset_output_buffer()

    def writePort(self, data):
        """
        Writes data to the serial port
        """
        if self.serial_port is not None:
            return self.serial_port.write(serial.to_bytes(data))

    def writePortClean(self, data):
        """
        Writes data to the serial port, but first clears the input buffer if
        there is data. This ensures that the next read will be aligned with
        the response from this write
        """

        if self.serial_port is not None:
            if self.numBytesInBuffer() > 0:
                self.flushPort()

            return self.serial_port.write(serial.to_bytes(data))

    def readPort(self, numBytes=None):
        if self.serial_port is not None:
            if numBytes is None:
                # No number of specified bytes to read. Read all available bytes
                return self.serial_port.read(self.serial_port.inWaiting())
            else:
                return self.serial_port.read(numBytes)

    def numBytesInBuffer(self):
        if self.serial_port is not None:
            return self.serial_port.inWaiting()
        return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('docs/history.rst', encoding='utf-8') as history_file:
    history = history_file.read()

with open('./requirements.txt', encoding='utf-8') as req_file:
    requirements = req_file.read()

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Micah Johnson",
    author_email='info@adventuredata.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Command line interface to the Realtime Adventure Data Lyte probe for measuring avalanche conditions",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='radicl',
    name='radicl',
    packages=find_packages(include=['radicl', 'radicl.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/adventuredata/radicl',
    version='0.7.2',
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'radicl = radicl.cli:main',
            'plotlyte = radicl.plotting:main',
            'lyte_hi_res = radicl.high_resolution:main',
            'plot_hi_res = radicl.plotting:plot_hi_res_cli',
        ],
    },
)



"""
This script shows the simplicity of using the radicl framework to access
settings information from the Lyte Probe.

This example the script collects the full calibrated signal from the probe
in addition to the depth data, allowing for high resolution profiles

Usage:  1. plug in the probe.
        2. Open a terminal
        3. python take_high_resolution_readings.py

"""

import argparse
from argparse import RawTextHelpFormatter
import pandas as pd
import json
import sys

from radicl import __version__
from radicl.interface import RADICL
from radicl.ui_tools import get_logger, exit_requested
from radicl.plotting import plot_hi_res
from radicl.gps import USBGPS


def build_high_resolution_data(cli, log):
    """
    Grabs the bottom sensors (sampled at the highest rate) then grabs the supporting sensors
    and pads with nans to fit into the same dataframe

    Args:
        cli: Instantiated Radicl() class
        log: Instantiated logger object

    Returns:
        result: Single data frame containing Force, NIR, Ambient NIR, Accel, Depth
    """
    # Grab the Raw data
    ts = cli.grab_data('rawsensor')

    # Grab relative, filtered barometer data
    depth = cli.grab_data('filtereddepth')

    # Grab Acceleration
    acc = cli.grab_data('rawacceleration')

    # Set the 0 point of depth to the Starting point or the snow Surface
    depth['depth'] = depth['filtereddepth'] - depth['filtereddepth'].min()

    # Invert Depth so bottom is negative max depth
    depth['depth'] = depth['depth'] - depth['depth'].max()
    depth = depth.drop(columns=['filtereddepth'])

    log.info("Barometer Depth achieved: {:0.1f} cm".format(abs(depth['depth'].max() - depth['depth'].min())))
    log.info("Barometer Samples: {:,}".format(len(depth.index)))
    log.info("Acceleration Samples: {:,}".format(len(acc.index)))
    log.info("Sensor Samples: {:,}".format(len(ts)))

    log.info("Infilling and interpolating dataset...")
    result = pd.merge_ordered(ts, depth, on='time')
    result = pd.merge_ordered(result, acc, on='time')
    result = result.interpolate(method='index')
    return result


def main():
    # Parse command line arguments
    hdr = 'Lyte Probe High Resolution DAQ Script v{}'.format(__version__)
    underline = '=' * len(hdr)
    hdr = '\n'.join([hdr, underline, ''])

    help_string = (
        '\nThis script is used to take high resolution hand driven measurements with'
        ' the Lyte probe. This script requires: \n'
        '* A USB connection to the probe\n'
        '* The user to press the probe button for start and stop\n\n'
        'To accurately take High resolution measurements this script automatically'
        ' downloads the following timeseries from the probe per measurement:\n'
        '* Force\n'
        '* Active NIR\n'
        '* Passive NIR\n'
        '* Depth\n'
        '* Acceleration in the line of pole\n\n'
        'NOTE: The depth and accelerometer timeseries are all recorded at a lower'
        ' sampling rate than the sensors in the tip so they are interpolated to'
        ' match the Force and NIR timeseries'
        'Provide calibration coefficients via the calibration.json file to plot'
        ' data calibrated (Note data is still stored raw)'
    )

    p = argparse.ArgumentParser(description='\n'.join([hdr, help_string]),
                                formatter_class=RawTextHelpFormatter)
    p.add_argument('-d', '--debug', dest='debug', action='store_true',
                   help="Debug flag will print out much more info")
    p.add_argument('-c', '--calibration', dest='calibration',
                   help='Path to a json containing any calibration coefficients for any of the sensors')
    p.add_argument('--version', action='version',
                   version='%(prog)s v{version}'.format(version=__version__))
    p.add_argument('--plot_time', default=10, type=int, help='Automatically close a plot after number of seconds')

    p.add_argument('--n_measurements', default=0, type=int, help='Number of measurements to take without asking to exit')
    args = p.parse_args()

    if args.calibration is not None:
        with open(args.calibration, 'r') as fp:
            calibration = json.load(fp)
    else:
        calibration = {'Sensor1': [1, 0], 'Sensor2': [1, 0], 'Sensor3': [1, 0], 'Sensor4': [1, 0]}

    # Start this scripts logging
    log = get_logger("RAD Hi-Res Script", debug=args.debug)

    log.info("Starting High Resolution DAQ Script")

    # Retrieve a connection to the probe
    cli = RADICL()

    # Look for a gps
    gps = USBGPS()

    # Keep count of measurements taken
    i = 0

    # Reset the probe in the event the probe was closed out without reset
    response = cli.probe.resetMeasurement()

    # Grab the probe sample rate
    SR = cli.probe.getSetting(setting_name='samplingrate')
    zpfo = cli.probe.getSetting(setting_name='zpfo')
    acc_range = cli.probe.getSetting(setting_name='accrange')

    finished = exit_requested()

    # Loop through each sensor and retrieve the calibration data
    while not finished:

        # take a measurement
        cli.listen_for_a_reading()

        # Collect and build the data
        ts = build_high_resolution_data(cli, log)

        meta = {"SAMPLE RATE": str(SR),
                "ZPFO": str(zpfo),
                "ACC. Range": str(acc_range),
                'Serial Num.': cli.probe.getProbeSerial()}

        # Attempt to get a fix, if no gps cnx then no location data is returned
        location = gps.get_fix()
        if location is not None:
            meta['Latitude'] = location[0]
            meta['Longitude'] = location[1]

        # Output the data to a datetime file
        cli.write_probe_data(ts, extra_meta=meta)

        # Plot the data
        plot_hi_res(df=ts, calibration_dict=calibration, timed_plot=args.plot_time)

        # Reset the probe / clear out the data
        cli.probe.resetMeasurement()

        i += 1
        log.info(f"{i} measurements taken this session")
        if i >= args.n_measurements:
            finished = exit_requested()

    log.info(f"{i} measurements taken this session")
    log.info("Exiting High Resolution DAQ Script")
    sys.exit()


if __name__ == '__main__':
    main()



import pytest
from time import sleep
from . import not_connected


@pytest.mark.skipif(not_connected, reason='probe not connected')
def test_getHWID(real_api):
    a = real_api.getHWID()
    assert type(a['data']) == int


@pytest.mark.skipif(not_connected, reason='probe not connected')
def test_MeasGetAccRange(real_api):
    a = real_api.MeasGetAccRange()
    a = int.from_bytes(a['data'], byteorder='little')
    assert (a >= 2) and (a <= 16)


@pytest.mark.skipif(not_connected, reason='probe not connected')
@pytest.mark.parametrize('value', [2, 4, 6, 8, 16])
def test_MeasSetAccRange(real_api, value):
    # Attempt to set the value
    r1 = real_api.MeasSetAccRange(value)
    sleep(0.1)

    # Retrieve it
    stored_val = real_api.MeasGetAccRange()
    stored_val = int.from_bytes(stored_val['data'], byteorder='little')

    assert stored_val == value
